import json
import os
import shutil
import logging
import yaml

from pathlib import Path
from git import Repo
from .train_model import TrainModel
from .evaluate_model import EvaluateModel
from .score_model import ScoreModel


class RepoManager(object):

    def __init__(self, base_path: str):
        self.base_path = base_path
        self.logger = logging.getLogger(__name__)

    def add_model(
        self,
        model_id: str,
        model_name: str,
        model_desc: str,
        template: str,
        base_path: str = None,
    ):

        base_path = self.base_path if base_path is None else base_path
        model_path = os.path.join(base_path, "model_definitions")

        if not os.path.isdir(model_path):
            raise ValueError("model directory {0} does not exist".format(model_path))

        with open(os.path.join(template, "model.json"), "r") as f:
            model = json.load(f)
            model["source_model_id"] = model["id"]
            model["id"] = model_id
            model["name"] = model_name
            model["description"] = model_desc

            name_path = model_name.strip().lower().replace(" ", "_")

            target_path = os.path.join(model_path, name_path)
            if os.path.isdir(target_path):
                raise ValueError(
                    "Model path {} already exists, please try another model name"
                    .format(name_path)
                )

            shutil.copytree(template, target_path)

            with open(os.path.join(target_path, "model.json"), "w") as t:
                json.dump(model, t, indent=4)

    def add_task(self, task: str, base_path: str = None):

        base_path = self.base_path if base_path is None else base_path
        task_path = os.path.join(base_path, "feature_engineering")

        if not os.path.exists(task_path):
            os.makedirs(task_path)

        for item in os.listdir(task):
            s = os.path.join(task, item)
            d = os.path.join(task_path, item)
            if os.path.isdir(s):
                shutil.copytree(s, d, dirs_exist_ok=True)
            else:
                shutil.copy2(s, d)

    @staticmethod
    def clone_repository(url, path, branch: str = "master"):
        repo = Repo.clone_from(url, path)
        repo.git.checkout(branch)

    def get_templates(self, source_path: str = None):

        source_path = (
            self.base_path if source_path is None else os.path.join(source_path, "")
        )
        model_path = os.path.join(source_path, "model_definitions")

        # if no models / not models repo, return empty.
        if not os.path.isdir(model_path):
            return {}

        templates = {}

        for model in os.listdir(model_path):
            model_metadata_dir = os.path.join(model_path, model)
            model_metadata_file = os.path.join(model_metadata_dir, "model.json")
            if os.path.isfile(model_metadata_file):
                with open(model_metadata_file, "r") as f:
                    model_metadata = json.load(f)
                    if not model_metadata["language"] in templates:
                        templates[model_metadata["language"]] = {}
                    templates[model_metadata["language"]][model_metadata["id"]] = [
                        model_metadata["name"],
                        model_metadata_dir,
                    ]

        return templates

    def get_task(self, source_path: str = None):

        source_path = (
            self.base_path if source_path is None else os.path.join(source_path, "")
        )
        task_path = os.path.join(source_path, "feature_engineering")

        # if no task / not models repo, return empty
        if not os.path.isdir(task_path):
            return ""

        # Check for the existence of 'task.py' and 'requirements.txt'
        required_files = ["task.py", "requirements.txt"]
        existing_files = os.listdir(task_path)
        for file in required_files:
            if file not in existing_files:
                return ""

        return task_path

    def init_model_directory(self, path: str = None):
        logging.info("Creating model directory")
        if path is None:
            path = os.path.join(os.path.abspath(os.getcwd()), "")

        self.logger.info("Creating model definitions")

        src = os.path.join(os.path.split(__file__)[0], "") + "metadata_files"
        src_files = os.listdir(src)
        for file_name in src_files:
            full_file_name = os.path.join(src, file_name)
            if os.path.isfile(full_file_name) and not os.path.exists(
                os.path.join(path, file_name)
            ):
                shutil.copy(full_file_name, path)

        Path(path + "model_definitions/").mkdir(parents=True, exist_ok=True)

        self.logger.info("model directory initialized at {0}".format(path))

    def read_repo_config(self):
        path = os.path.join(self.base_path, ".aoa/config.yaml")
        if os.path.exists(path):
            with open(path, "r") as handle:
                return yaml.safe_load(handle)

        self.logger.warning("Aoa repo config doesn't exist")
        return None

    def write_repo_config(self, config, path=None):
        path = path if path else self.base_path
        config_dir = os.path.join(path, ".aoa")
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        config_file = "{}/config.yaml".format(config_dir)

        with open(config_file, "w+") as f:
            yaml.dump(config, f, default_flow_style=False)

    def repo_config_exists(self, repo_path=None):
        path = repo_path if repo_path else self.base_path
        return Path(os.path.join(path, ".aoa")).is_file()

    def train(self, model_id: str, data_conf: dict):
        trainer = TrainModel(self)
        trainer.train_model_local(model_id=model_id, rendered_dataset=data_conf)

    def evaluate(self, model_id: str, data_conf: dict):
        evaluator = EvaluateModel(self)
        evaluator.evaluate_model_local(model_id=model_id, rendered_dataset=data_conf)

    def batch_score_model_local(self, model_id: str, data_conf: dict):
        scorer = ScoreModel(self)
        scorer.batch_score_model_local(model_id=model_id, rendered_dataset=data_conf)



import abc
from aoa.api.base_api import BaseApi
from aoa.api.api_iterator import ApiIterator


class IteratorBaseApi(BaseApi):
    __metaclass__ = abc.ABCMeta

    def __init__(self, aoa_client=None, iterator_projection=None, show_archived=True):
        super().__init__(aoa_client=aoa_client)

        self.iterator_projection = iterator_projection

        if show_archived:
            self.iterator_func = self.find_all
        else:
            self.iterator_func = self.find_by_archived

    def __iter__(self):
        return ApiIterator(
            api_func=self.iterator_func,
            entities=list(filter(str.strip, self.path.split("/")))[-1],
            projection=self.iterator_projection,
        )

    def __len__(self):
        return self.find_all()["page"]["totalElements"]



from __future__ import absolute_import
from typing import Dict

from aoa.api.iterator_base_api import IteratorBaseApi


class TrainedModelApi(IteratorBaseApi):

    path = "/api/trainedModels"
    type = "TRAINED_MODEL"

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.aoa_client.select_header_accept([
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]),
        ]

        return self.generate_params(header_vars, header_vals)

    def find_dataset(self, trained_model_id: str, projection: str = None):
        """
        returns dataset of a trained model

        Parameters:
           trained_model_id (str): trained model id(uuid)
           projection (str): projection type

        Returns:
            (dict): dataset
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/{trained_model_id}/dataset",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_events(self, trained_model_id: str, projection: str = None):
        """
        returns trained model events

        Parameters:
           trained_model_id (str): trained model id(uuid)
           projection (str): projection type

        Returns:
            (dict): events of trained model
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/{trained_model_id}/events",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_model_id(self, model_id: str, projection: str = None):
        """
        returns a trained models by model id

        Parameters:
           model_id (str): model id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): trained models
        """
        query_vars = ["modelId", "projection"]
        query_vals = [model_id, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByModelId",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_model_id_and_status(
        self, model_id: str, status: str, projection: str = None
    ):
        """
        returns a trained models by model id

        Parameters:
           model_id (str): model id(uuid) to find
           status (str): model status
           projection (str): projection type

        Returns:
            (dict): trained models
        """
        query_vars = ["modelId", "status", "projection"]
        query_vals = [model_id, status, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByModelIdAndStatus",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def evaluate(self, trained_model_id: str, evaluation_request: Dict[str, str]):
        """
        evaluate a model

        Parameters:
           trained_model_id (str): trained model id(uuid) to evaluate
           evaluation_request (dict): request to evaluate trained model

        Returns:
            (dict): job
        """
        self.required_params(["datasetId"], evaluation_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{trained_model_id}/evaluate",
            header_params=self._get_header_params(),
            query_params={},
            body=evaluation_request,
        )

    def approve(self, trained_model_id: str, comments: str):
        """
        approve a trained model
        :param trained_model_id:  model version
        :param comments: approval comments
        :return:
        """
        approve_request = dict()
        approve_request["comments"] = comments

        return self.aoa_client.post_request(
            path=f"{self.path}/{trained_model_id}/approve",
            header_params=self._get_header_params(),
            query_params={},
            body=approve_request,
        )

    def reject(self, trained_model_id: str, comments: str):
        """
        reject a trained model
        :param trained_model_id:  model version
        :param comments: approval comments
        :return:
        """
        reject_request = dict()
        reject_request["comments"] = comments

        return self.aoa_client.post_request(
            path=f"{self.path}/{trained_model_id}/reject",
            header_params=self._get_header_params(),
            query_params={},
            body=reject_request,
        )

    def deploy(self, trained_model_id: str, deploy_request: dict):
        """
        deploy a trained model
        :param trained_model_id:  model version
        :param deploy_request: deployment request
        :return:
        """
        self.required_params(["engineType"], deploy_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{trained_model_id}/deploy",
            header_params=self._get_header_params(),
            query_params={},
            body=deploy_request,
        )

    def retire(self, trained_model_id: str, retire_request: dict):
        """
        retire a trained model
        :param trained_model_id:  model version
        :param retire_request: retire request
        :return:
        """
        self.required_params(["deploymentId"], retire_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{trained_model_id}/retire",
            header_params=self._get_header_params(),
            query_params={},
            body=retire_request,
        )



from __future__ import absolute_import

from aoa.api.iterator_base_api import IteratorBaseApi


class JobEventApi(IteratorBaseApi):

    path = "/api/jobEvents"
    type = "JOB_EVENT"

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.aoa_client.select_header_accept([
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]),
        ]

        return self.generate_params(header_vars, header_vals)

    def find_all(
        self,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns all job events

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): all job events
        """
        query_vars = ["projection", "page", "sort", "size", "sort"]
        query_vals = [projection, page, sort, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_id(self, job_event_id: str, projection: str = None):
        """
        returns a job event

        Parameters:
           job_event_id (str): job event id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): job event
        """
        query_vars = ["id", "projection"]
        query_vals = [job_event_id, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findById",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_job_id(
        self,
        job_id: str,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns events of a job

        Parameters:
           job_id (str): job id(uuid) to find events
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): job events
        """
        query_vars = ["jobId", "projection", "page", "sort", "size", "sort"]
        query_vals = [job_id, projection, page, sort, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByJobId",
            header_params=self._get_header_params(),
            query_params=query_params,
        )



from .base_model import BaseModel
from ..context.model_context import ModelContext, DatasetInfo
import json
import os
import importlib
import logging
import shutil
import sys


class ScoreModel(BaseModel):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def batch_score_model_local(
        self, model_id: str = None, base_path: str = None, rendered_dataset: dict = None
    ):
        base_path = (
            self.repo_manager.model_catalog_path
            if base_path is None
            else os.path.join(base_path, "")
        )
        model_definitions_path = base_path + "model_definitions/"

        if not os.path.exists(model_definitions_path):
            raise ValueError(
                "model directory {0} does not exist".format(model_definitions_path)
            )

        model_artefacts_path = ".artefacts/{}".format(model_id)

        if not os.path.exists("{}/output".format(model_artefacts_path)):
            raise ValueError("You must run training before trying to run scoring.")

        try:
            self.__cleanup()

            os.makedirs("./artifacts")

            os.symlink(
                "../{}/output/".format(model_artefacts_path),
                "./artifacts/input",
                target_is_directory=True,
            )
            os.symlink(
                "../{}/output/".format(model_artefacts_path),
                "./artifacts/output",
                target_is_directory=True,
            )

            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )

            with open(model_dir + "/model.json", "r") as f:
                model_definition = json.load(f)

            with open(model_dir + "/config.json", "r") as f:
                model_conf = json.load(f)

            self.logger.info("Loading and executing model code")

            cli_model_kargs = self._BaseModel__get_model_varargs(model_id)

            context = ModelContext(
                dataset_info=DatasetInfo.from_dict(rendered_dataset),
                hyperparams=model_conf["hyperParameters"],
                artifact_input_path="artifacts/input",
                artifact_output_path="artifacts/output",
                **cli_model_kargs,
            )

            engine = self._BaseModel__get_engine(model_definition, "deployment")
            if engine == "python" or engine == "pyspark":
                sys.path.append(model_dir)
                scoring = importlib.import_module(".scoring", package="model_modules")

                scoring.score(
                    context=context,
                    data_conf=rendered_dataset,
                    model_conf=model_conf,
                    **cli_model_kargs,
                )

            elif engine == "sql":
                raise Exception("not supported")

            elif engine == "R":
                self._BaseModel__run_r_model(
                    model_id, model_dir, rendered_dataset, "score.batch"
                )

            else:
                raise Exception(
                    "Unsupported language: {}".format(model_definition["language"])
                )

            self.__cleanup()
        except ModuleNotFoundError:
            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )
            self.__cleanup()
            self.logger.error(
                "Missing required python module, try running following command first:"
            )
            self.logger.error(
                "pip install -r {}/model_modules/requirements.txt".format(model_dir)
            )
            raise
        except:
            self.__cleanup()
            self.logger.exception("Exception running model code")
            raise

    def __cleanup(self):
        if os.path.exists("./artifacts"):
            shutil.rmtree("./artifacts")
        try:
            os.remove("./models")
        except FileNotFoundError:
            self.logger.debug("Nothing to remove, folder './models' does not exists.")



from __future__ import absolute_import
from typing import Dict

from aoa.api.iterator_base_api import IteratorBaseApi


class ModelApi(IteratorBaseApi):

    path = "/api/models"
    type = "MODEL"

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.aoa_client.select_header_accept([
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]),
        ]

        return self.generate_params(header_vars, header_vals)

    def find_by_source_id(self, source_model_id: str, projection: str = None):
        """
        returns a model by source model id taken from git repo

        Parameters:
           source_model_id (str): source model id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): model
        """
        query_vars = ["sourceId", "projection"]
        query_vals = [source_model_id, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            f"{self.path}/search/findBySourceId",
            self._get_header_params(),
            query_params,
        )

    def find_all_commits(self, model_id: str, projection: str = None):
        """
        returns model commits

        Parameters:
           model_id (str): model id(uuid) for commits
           projection (str): projection type

        Returns:
            (dict): model commits
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            f"{self.path}/{model_id}/commits", self._get_header_params(), query_params
        )

    def diff_commits(
        self, model_id: str, commit_id1: str, commit_id2: str, projection: str = None
    ):
        """
        returns difference between model commits

        Parameters:
           model_id (str): model id(uuid)
           commit_id1 (str): id of commit to compare
           commit_id2 (str): id of commit to compare
           projection (str): projection type

        Returns:
            (str): difference between model commits
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            f"{self.path}/{model_id}/diff/{commit_id1}/{commit_id2}/",
            self._get_header_params(),
            query_params,
        )

    def save(self, model: Dict[str, str]):
        """
        register a dataset

        Parameters:
           model (dict): external model to register

        Returns:
            (dict): model
        """

        return self.aoa_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=model,
        )

    def train(self, model_id: str, training_request: Dict[str, str]):
        """
        train a model

        Parameters:
            model_id (str): model id(uuid)
            training_request (dict): request to train model

        Returns:
            (dict): job
        """
        self.required_params(["datasetId"], training_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{model_id}/train",
            header_params=self._get_header_params(),
            query_params={},
            body=training_request,
        )

    def import_byom(self, model_id: str, import_request: Dict[str, str]):
        """
        import a model version

        Parameters:
            model_id (str): model id(uuid)
            import_request (dict): request to import model version

        Returns:
            (dict): job
        """
        self.required_params(["artefactImportId", "externalId"], import_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{model_id}/import",
            header_params=self._get_header_params(),
            query_params={},
            body=import_request,
        )



import json
import pandas as pd
from typing import Dict
import logging
from ..util.connections import execute_sql
from teradataml import DataFrame

logger = logging.getLogger(__name__)

byom_query = """
CREATE TABLE {} (
      model_version VARCHAR(255),
      model_id VARCHAR(255),
      model_type VARCHAR(255),
      project_id VARCHAR(255),
      deployed_at TIMESTAMP(6) DEFAULT CURRENT_TIMESTAMP(6),
      model BLOB(2097088000))
UNIQUE PRIMARY INDEX (model_version);
"""

merge_query = """
MERGE {} target
     USING {} source
       ON target.column_name = source.column_name
     WHEN MATCHED THEN
       UPD SET stats = source.stats, column_type = source.column_type, update_ts = source.update_ts
     WHEN NOT MATCHED THEN
       INS (source.column_name, source.column_type, source.stats, source.update_ts);
"""
temp_table = "aoa_stats_temp"


def save_feature_stats(features_table: str, feature_type: str, stats: Dict) -> None:
    cvt_query = (
        f"CREATE VOLATILE TABLE {temp_table} AS {features_table} WITH NO DATA ON COMMIT"
        " PRESERVE ROWS;"
    )
    ins_query = (
        f"INS {temp_table} (column_name,column_type,stats,update_ts)"
        " VALUES(?,?,?,CURRENT_TIMESTAMP);"
    )
    m_query = merge_query.format(features_table, temp_table)
    dt_query = f"DROP TABLE {temp_table};"

    logger.debug(cvt_query)
    execute_sql(cvt_query)
    logger.debug(ins_query)
    execute_sql(ins_query, [[f, feature_type, json.dumps(stats[f])] for f in stats])
    logger.debug(m_query)
    execute_sql(m_query)
    logger.debug(dt_query)
    execute_sql(dt_query)


def get_feature_stats_summary(features_table: str) -> Dict:
    try:
        fs = DataFrame.from_query(
            f"SEL column_name, column_type FROM {features_table}"
        ).to_pandas(all_rows=True)
        fs = fs.reset_index().drop(
            fs.columns.difference(["column_name", "column_type"]), axis=1
        )
        fs = fs.set_index("column_name")
        return pd.Series(fs.column_type).to_dict()
    except Exception:
        logger.warning("Couldn't read statistics metadata, assuming it's empty")
        return {}


def get_feature_stats(features_table: str, feature_type: str) -> Dict:
    try:
        fs = DataFrame.from_query(
            f"SEL * FROM {features_table} WHERE column_type='{feature_type}'"
        ).to_pandas(all_rows=True)
        fs = fs.reset_index().drop(
            fs.columns.difference(["column_name", "stats"]), axis=1
        )
        fs = fs.set_index("column_name")
        fs = pd.Series(fs.stats).to_dict()
        return {k: json.loads(fs[k]) for k in fs}
    except Exception:
        logger.warning("Couldn't read statistics metadata, assuming it's empty")
        return {}


def get_features_stats_metadata_ct_query(tablename: str, volatile: bool = False) -> str:
    # It's crucial that we use column_name as a unique index here, because UI currently doesn't allow to set a feature type
    # If the user wants the same column to play as both continuous and categorical features, she must provide a copy of the column with a different name
    ct_query = """
  CREATE {} (
      column_name VARCHAR(128), 
      column_type VARCHAR(128),
      stats JSON, 
      update_ts TIMESTAMP)
  UNIQUE PRIMARY INDEX ( column_name ){}
  """
    if volatile:
        return ct_query.format(
            f"VOLATILE TABLE {tablename}", " ON COMMIT PRESERVE ROWS;"
        )
    else:
        return ct_query.format(f"TABLE {tablename}", ";")


def create_features_stats_table(features_table: str, volatile: bool = False) -> None:
    query = get_features_stats_metadata_ct_query(features_table, volatile)
    logger.debug(query)
    execute_sql(query)



from aoa import ModelContext


def save_plot(title: str, dpi: int = 500, context: ModelContext = None):
    """
    This assumes the plot has already been rendered and we only need to save it.
    :param title: the title of the plot and the file name to save as (spaces replaced with _ and all lower)
    :param dpi: the dpi setting
    :param context: ModelContext
    :return: None
    """
    import matplotlib.pyplot as plt

    plt.title(title)
    fig = plt.gcf()
    title = title.replace(" ", "_").lower()

    filename = f"artifacts/output/{title}"
    if context:
        filename = f"{context.artifact_output_path}/{title}"

    fig.savefig(filename, dpi=dpi)
    plt.clf()



from teradataml import create_context, get_context, configure

import os
import logging

logger = logging.getLogger(__name__)


def aoa_create_context(database: str = None):
    """
    Creates a teradataml context if one does not already exist.
    Most users should not need to understand how we pass the environment variables etc. for dataset connections. This
    provides a way to achieve that and also allow them to work within a notebook for example where a context is already
    present.

    We create the connection based on the following environment variables which are configured automatically by the
    aoa based on the dataset connection selected:

        AOA_CONN_HOST
        AOA_CONN_USERNAME
        AOA_CONN_PASSWORD
        AOA_CONN_LOG_MECH
        AOA_CONN_DATABASE
        AOA_VAL_INSTALL_DB
        AOA_BYOM_INSTALL_DB

    :param database: default database override
    :return: None
    """
    if get_context() is None:
        if not database:
            database = os.getenv("AOA_CONN_DATABASE")

        host = os.environ["AOA_CONN_HOST"]
        logmech = os.getenv("AOA_CONN_LOG_MECH", "TDNEGO")
        username = os.environ["AOA_CONN_USERNAME"]
        password = os.environ["AOA_CONN_PASSWORD"]

        if database:
            logger.debug(f"Configuring temp database for tables/views to {database}")
            configure.temp_table_database = database
            configure.temp_view_database = database

        configure.val_install_location = os.environ.get("AOA_VAL_INSTALL_DB", "VAL")
        configure.byom_install_location = os.environ.get("AOA_BYOM_INSTALL_DB", "MLDB")

        logger.debug(
            f"Connecting to {host} on database {database} using logmech {logmech} as"
            f" {username}"
        )

        create_context(
            host=host,
            database=database,
            username=username,
            password=password,
            logmech=logmech,
        )
        os.unsetenv("AOA_CONN_USERNAME")
        os.unsetenv("AOA_CONN_PASSWORD")

        from aoa import __version__

        execute_sql(f"""
        SET QUERY_BAND = 'appVersion={__version__};appName=VMO;appFunc=python;org=teradata-internal-telem;' FOR SESSION VOLATILE
        """)

    else:
        logger.info("teradataml context already exists. Skipping create_context.")


def execute_sql_in_context(context, statement, parameters=None):
    from teradataml.common.exceptions import TeradataMlException
    from teradataml.common.messages import Messages
    from teradataml.common.messagecodes import MessageCodes

    if context is not None:
        tdsql_con = context.raw_connection().driver_connection
        cursor = tdsql_con.cursor()
        return cursor.execute(statement, parameters)
    else:
        raise TeradataMlException(
            Messages.get_message(MessageCodes.INVALID_CONTEXT_CONNECTION),
            MessageCodes.INVALID_CONTEXT_CONNECTION,
        )


def execute_sql(statement, parameters=None):
    return execute_sql_in_context(get_context(), statement, parameters)


def tmo_create_context(database: str = None):
    """
    Creates a teradataml context if one does not already exist.
    Most users should not need to understand how we pass the environment variables etc. for dataset connections. This
    provides a way to achieve that and also allow them to work within a notebook for example where a context is already
    present.

    We create the connection based on the following environment variables which are configured automatically by the
    aoa based on the dataset connection selected:

        AOA_CONN_HOST
        AOA_CONN_USERNAME
        AOA_CONN_PASSWORD
        AOA_CONN_LOG_MECH
        AOA_CONN_DATABASE
        AOA_VAL_INSTALL_DB
        AOA_BYOM_INSTALL_DB

    :param database: default database override
    :return: None
    """
    return aoa_create_context(database)



import base64, secrets, datetime

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from .TJEncryptPassword import (
    createPasswordEncryptionKeyFile,
    createEncryptedPasswordFile,
    decryptPassword,
)

backend = default_backend()

default_conf = {
    "iterations": 65536,
    "bytes_key_length": 32,
    "bytes_salt_length": 16,
    "bytes_iv_length": 16,
    "bytes_auth_tag_length": 16,
}


def _derive_key(password, salt, iterations, length):
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=length,
        salt=salt,
        iterations=iterations,
        backend=backend,
    )
    return kdf.derive(password)


def aes_gcm_encrypt(message: str, password: str, conf: dict = default_conf) -> str:
    """
    Encrypts text for AOA core for e.g connection credentials. It is AES 256 bit HMAC

    :param message: the text to encrypt
    :param password: the token/key to use for decryption (no spaces allowed)
    :param conf: if not specified will use the default conf of
            default_conf = {
                "iterations": 65536,
                "bytes_key_length": 32,
                "bytes_salt_length": 16,
                "bytes_iv_length": 16,
                "bytes_auth_tag_length": 16
            }
    :return: the encrypted message
    """
    auth_tag = secrets.token_bytes(conf["bytes_auth_tag_length"])
    iv = secrets.token_bytes(conf["bytes_iv_length"])
    salt = secrets.token_bytes(conf["bytes_salt_length"])
    key = _derive_key(
        password.encode(), salt, conf["iterations"], conf["bytes_key_length"]
    )
    algorithm = algorithms.AES(key)
    cipher = Cipher(algorithm, modes.GCM(iv), backend=backend)
    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(message.encode())

    return base64.urlsafe_b64encode(iv + salt + ciphertext + auth_tag).decode("utf-8")


def aes_gcm_decrypt(message: str, password: str, conf: dict = default_conf) -> str:
    """
    Decrypt text which was previously encrypted by the AOA core for e.g connection credentials. It is AES 256 bit HMAC

    :param message: the text to decrypt
    :param password: the token/key to use for decryption
    :param conf: if not specified will use the default conf of
            default_conf = {
                "iterations": 65536,
                "bytes_key_length": 32,
                "bytes_salt_length": 16,
                "bytes_iv_length": 16,
                "bytes_auth_tag_length": 16
            }
    :return: decrypted text
    """
    data = base64.urlsafe_b64decode(message.encode())
    (iv, salt, ciphertext, auth_tag) = (
        data[: conf["bytes_iv_length"]],
        data[
            conf["bytes_iv_length"] : conf["bytes_iv_length"]
            + conf["bytes_salt_length"]
        ],
        data[
            conf["bytes_iv_length"]
            + conf["bytes_salt_length"] : -conf["bytes_auth_tag_length"]
        ],
        data[-conf["bytes_auth_tag_length"] :],
    )
    key = _derive_key(
        password.encode(),
        salt,
        iterations=conf["iterations"],
        length=conf["bytes_key_length"],
    )
    algorithm = algorithms.AES(key)
    cipher = Cipher(algorithm, modes.GCM(iv, salt), backend=backend)
    decryptor = cipher.decryptor()
    decrypted = decryptor.update(ciphertext)

    return decrypted.decode("utf-8")


def td_encrypt_password(
    password=None,
    pass_key_filename="~/.aoa/userKey.properties",
    enc_pass_filename="~/.aoa/userPass.properties",
    transformation="AES/CBC/PKCS5Padding",
    key_size_in_bits=256,
    mac="HmacSHA256",
):
    """
    Encrypts a password, saves the encryption key in one file, and saves the encrypted password in a second file.
    This is done using the Stored Password Protection format.
    e.g. "ENCRYPTED_PASSWORD(file:PasswordEncryptionKeyFileName,file:EncryptedPasswordFileName)"

    :param password: password to encrypt
    :param pass_key_filename: Specifies the location of the generated password encryption file
    :param enc_pass_filename: Specifies the location of the generated encrypted password file
    :param transformation: Specifies the transformation in the form Algorithm/Mode/Padding
    :param key_size_in_bits: the encrypted key size in bits
    :param mac: Specifies the message authentication code (MAC) algorithm HmacSHA1 or HmacSHA256
    :return: the encrypted password
    """

    transformation_parts = transformation.split("/")
    if len(transformation_parts) != 3:
        raise ValueError("Invalid transformation " + transformation)

    algorithm = transformation_parts[0]
    mode = transformation_parts[1]
    padding = transformation_parts[2]

    if algorithm not in ["DES", "DESede", "AES"]:
        raise ValueError("Unknown algorithm " + algorithm)

    if mode not in ["CBC", "CFB", "OFB"]:
        raise ValueError("Unknown mode " + mode)

    if padding not in ["PKCS5Padding", "NoPadding"]:
        raise ValueError("Unknown padding " + padding)

    if mac not in ["HmacSHA1", "HmacSHA256"]:
        raise ValueError("Unknown MAC algorithm " + mac)

    if not password:
        raise ValueError("Password cannot be zero length")

    password = password.encode().decode("unicode_escape")

    key_size_in_bits = int(key_size_in_bits)
    match = str(datetime.datetime.now())

    aby_key, aby_mac_key = createPasswordEncryptionKeyFile(
        transformation,
        algorithm,
        mode,
        padding,
        key_size_in_bits,
        match,
        mac,
        pass_key_filename,
    )

    createEncryptedPasswordFile(
        transformation,
        algorithm,
        mode,
        padding,
        match,
        aby_key,
        mac,
        aby_mac_key,
        enc_pass_filename,
        password,
    )

    return "ENCRYPTED_PASSWORD(file:{},file:{})".format(
        pass_key_filename, enc_pass_filename
    )


def td_decrypt_password(password=None):
    """
    Decrypts a password, from the encryption key file and the encrypted password file.
    This is done using the Stored Password Protection format.

    :param password: Specifies the encrypted password string
    :return: the decrypted password
    """

    if not password:
        raise ValueError("Password cannot be zero length")

    import re

    password_files = re.match("^ENCRYPTED_PASSWORD\\(file:(.*),file:(.*)\\)$", password)

    return decryptPassword(password_files[1], password_files[2])



__version__ = "7.1.1"

# import client
from aoa.api_client import AoaClient

# import APIs into api package
from aoa.api.dataset_api import DatasetApi
from aoa.api.dataset_template_api import DatasetTemplateApi
from aoa.api.dataset_connection_api import DatasetConnectionApi
from aoa.api.model_api import ModelApi
from aoa.api.project_api import ProjectApi
from aoa.api.trained_model_api import TrainedModelApi
from aoa.api.trained_model_event_api import TrainedModelEventApi
from aoa.api.trained_model_artefacts_api import TrainedModelArtefactsApi
from aoa.api.job_api import JobApi
from aoa.api.job_event_api import JobEventApi
from aoa.api.deployment_api import DeploymentApi
from aoa.api.api_iterator import ApiIterator
from aoa.api.message_api import MessageApi
from aoa.api.user_attributes_api import UserAttributesApi
from aoa.api.feature_engineering_api import FeatureEngineeringApi

# import repo into api package
from aoa.cli.repo_manager import RepoManager
from aoa.cli.evaluate_model import EvaluateModel
from aoa.cli.score_model import ScoreModel
from aoa.cli.train_model import TrainModel
from aoa.cli.base_model import BaseModel
from aoa.cli.run_task import RunTask
from aoa.cli.base_task import BaseTask

from aoa.context.model_context import *
from aoa.util import *
from aoa.stats.stats import *
from aoa.stats.store import *



from .base_model import BaseModel
from ..context.model_context import ModelContext, DatasetInfo
import json
import os
import importlib
import logging
import shutil
import sys


class EvaluateModel(BaseModel):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def evaluate_model_local(
        self, model_id: str, base_path: str, rendered_dataset: dict
    ):

        base_path = (
            self.repo_manager.model_catalog_path
            if base_path is None
            else os.path.join(base_path, "")
        )
        model_definitions_path = base_path + "model_definitions/"

        if not os.path.exists(model_definitions_path):
            raise ValueError(
                "model directory {0} does not exist".format(model_definitions_path)
            )

        model_artefacts_path = ".artefacts/{}".format(model_id)
        model_evaluation_path = "{}/evaluation".format(model_artefacts_path)

        if not os.path.exists("{}/output".format(model_artefacts_path)):
            raise ValueError("You must run training before trying to run evaluation.")

        if os.path.exists(model_evaluation_path):
            self.logger.debug(
                "Cleaning local model evaluation path: {}".format(model_evaluation_path)
            )
            shutil.rmtree(model_evaluation_path)

        os.makedirs("{}/output".format(model_evaluation_path))

        try:
            self.__cleanup()

            os.makedirs("./artifacts")

            os.symlink(
                "../{}/output/".format(model_artefacts_path),
                "./artifacts/input",
                target_is_directory=True,
            )
            os.symlink(
                "../{}/evaluation/output/".format(model_artefacts_path),
                "./artifacts/output",
                target_is_directory=True,
            )

            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )

            with open(model_dir + "/model.json", "r") as f:
                model_definition = json.load(f)

            with open(model_dir + "/config.json", "r") as f:
                model_conf = json.load(f)

            self.logger.info("Loading and executing model code")

            cli_model_kargs = self._BaseModel__get_model_varargs(model_id)

            context = ModelContext(
                dataset_info=DatasetInfo.from_dict(rendered_dataset),
                hyperparams=model_conf["hyperParameters"],
                artifact_input_path="artifacts/input",
                artifact_output_path="artifacts/output",
                **cli_model_kargs,
            )

            engine = self._BaseModel__get_engine(model_definition, "evaluation")
            if engine == "python":
                sys.path.append(model_dir)
                if os.path.isfile("{}/model_modules/evaluation.py".format(model_dir)):
                    evaluation = importlib.import_module(
                        ".evaluation", package="model_modules"
                    )
                else:
                    logging.debug(
                        "No evaluation.py found. Using scoring.py -> evaluate"
                    )
                    evaluation = importlib.import_module(
                        ".scoring", package="model_modules"
                    )

                evaluation.evaluate(
                    context=context,
                    data_conf=rendered_dataset,
                    model_conf=model_conf,
                    **cli_model_kargs,
                )

            elif engine == "sql":
                self.__evaluate_sql(
                    context, model_dir, rendered_dataset, model_conf, **cli_model_kargs
                )

            elif engine == "R":
                self._BaseModel__run_r_model(
                    model_id, model_dir, rendered_dataset, "evaluate"
                )

            else:
                raise Exception(
                    "Unsupported language: {}".format(model_definition["language"])
                )

            if os.path.exists("{}/evaluation/output/".format(model_artefacts_path)):
                self.logger.info(
                    "Artefacts can be found in: {}/evaluation/output/".format(
                        model_artefacts_path
                    )
                )
            else:
                self.logger.info(
                    "Artefacts can be found in: {}".format(model_artefacts_path)
                )

            if os.path.exists(
                "{}/evaluation/output/metrics.json".format(model_artefacts_path)
            ):
                self.logger.info(
                    "Evaluation metrics can be found in:"
                    f" {model_artefacts_path}/evaluation/output/metrics.json"
                )

            if os.path.exists("{}/evaluation.json".format(model_artefacts_path)):
                self.logger.info(
                    "Evaluation metrics can be found in: {}/evaluation.json".format(
                        model_artefacts_path
                    )
                )

            self.__cleanup()
        except ModuleNotFoundError:
            model_dir = model_definitions_path + BaseModel.get_model_folder(
                model_definitions_path, model_id
            )
            self.__cleanup()
            self.logger.error(
                "Missing required python module, try running following command first:"
            )
            self.logger.error(
                "pip install -r {}/model_modules/requirements.txt".format(model_dir)
            )
            raise
        except:
            self.__cleanup()
            self.logger.exception("Exception running model code")
            raise

    def __cleanup(self):
        if os.path.exists("./artifacts"):
            shutil.rmtree("./artifacts")
        try:
            os.remove("./models")
        except FileNotFoundError:
            self.logger.debug("Nothing to remove, folder './models' does not exists.")

    def __evaluate_sql(self, context, model_dir, data_conf, model_conf, **kwargs):
        from aoa.util import aoa_create_context
        from teradataml import remove_context, DataFrame

        self.logger.info("Starting evaluation...")

        aoa_create_context()

        sql_file = model_dir + "/model_modules/evaluation.sql"
        jinja_ctx = {
            "context": context,
            "data_conf": data_conf,
            "model_conf": model_conf,
            "model_table": kwargs.get("model_table"),
            "model_version": kwargs.get("model_version"),
            "model_id": kwargs.get("model_id"),
        }

        self._BaseModel__execute_sql_script(sql_file, jinja_ctx)

        self.logger.info("Finished evaluation")

        stats = DataFrame(data_conf["metrics_table"]).to_pandas(all_rows=True)
        metrics = dict(zip(stats.key, stats.value))

        with open("models/evaluation.json", "w+") as f:
            json.dump(metrics, f)

        self.logger.info("Saved metrics")

        remove_context()



import logging
import json
import yaml
import os
from jinja2 import Template


class BaseModel(object):

    def __init__(self, repo_manager):
        self.repo_manager = repo_manager

    @staticmethod
    def get_model_ids(model_path, rtn_val=False):
        catalog = {}
        index = 0
        model_ids = "Use one of the following models\n"

        for model_folder in os.listdir(model_path):
            if os.path.exists(model_path + model_folder + "/model.json"):
                with open(model_path + model_folder + "/model.json", "r") as f:
                    model_definition = json.load(f)
                    catalog[index] = model_definition
                    index += 1

        for key in catalog:
            model_ids += "{1}: {0}\n".format(catalog[key]["id"], catalog[key]["name"])

        if rtn_val:
            return catalog

        raise ValueError(model_ids)

    @staticmethod
    def get_model_folders(model_path, rtn_val=False):
        catalog = {}
        index = 0
        model_ids = "Use one of the following models\n"

        for model_folder in os.listdir(model_path):
            if os.path.exists(model_path + model_folder + "/model.json"):
                with open(model_path + model_folder + "/model.json", "r") as f:
                    model_definition = json.load(f)
                    catalog[index] = model_definition
                    index += 1

        for key in catalog:
            model_ids += "{1}: {0}\n".format(model_folder, catalog[key]["name"])

        if rtn_val:
            return catalog

        raise ValueError(model_ids)

    @staticmethod
    def get_model_folder(model_path: str, model_id: str):
        for model_folder in os.listdir(model_path):
            if os.path.exists(model_path + model_folder + "/model.json"):
                with open(model_path + model_folder + "/model.json", "r") as f:
                    model_definition = json.load(f)
                    if model_definition["id"] == model_id:
                        return model_folder

        raise ValueError("Could not find model path for model_id: {}".format(model_id))

    def __get_model_varargs(self, model_id):
        return {
            "model_id": model_id,
            "model_version": "cli",
            "model_table": "aoa_models_cli",
            "project_id": self.__get_project_id(),
            "job_id": "cli",
        }

    def __get_project_id(self):
        path = os.path.join(self.repo_manager.base_path, ".aoa/config.yaml")
        with open(path, "r") as handle:
            return yaml.safe_load(handle)["project_id"]

    def __template_sql_script(self, filename, jinja_ctx):
        with open(filename) as f:
            template = Template(f.read(), autoescape=True)

        return template.render(jinja_ctx)

    def __execute_sql_script(self, filename, jinja_ctx):
        from ..util.connections import execute_sql

        script = self.__template_sql_script(filename, jinja_ctx)

        stms = script.split(";")

        for stm in stms:
            stm = stm.strip()
            if stm:
                logging.info("Executing statement: {}".format(stm))

                try:
                    execute_sql(stm)
                except Exception as e:
                    if stm.startswith("DROP"):
                        logging.warning("Ignoring DROP statement exception")
                    else:
                        raise e

    def __get_engine(self, model_definition, mode):
        if "automation" in model_definition:

            if (
                mode in model_definition["automation"]
                and "engine" in model_definition["automation"][mode]
            ):
                return model_definition["automation"][mode]["engine"]

            # legacy
            if "trainingEngine" in model_definition["automation"]:
                return model_definition["automation"]["trainingEngine"]

        return model_definition["language"]

    def __run_r_model(self, model_id, model_dir, data_conf, mode):
        import tempfile
        import subprocess

        with tempfile.NamedTemporaryFile(delete=False) as fp:
            fp.write(json.dumps(data_conf).encode())

        pkg_path, _ = os.path.split(__file__)
        cmd = (
            f"{pkg_path}/run_model.R"
            f" {model_id} {self.__get_project_id()} {mode} {fp.name} {model_dir}"
        )
        subprocess.check_call(cmd, shell=True)



from __future__ import absolute_import
from typing import List
from aoa.api.base_api import BaseApi

import requests
import uuid
import os

SUPPORTED_BYOM_FORMATS = ["PMML", "ONNX", "H2O", "H2O_DAI", "DATAROBOT", "DATAIKU"]


class TrainedModelArtefactsApi(BaseApi):

    path = "/api/trainedModels/{}/artefacts"
    type = "TRAINED_MODEL_ARTEFACTS"

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.json_type,
        ]

        return self.generate_params(header_vars, header_vals)

    def list_artefacts(self, trained_model_id: uuid):
        """
        returns all trained models

        Parameters:
           trained_model_id (uuid): Trained Model Id

        Returns:
            (list): all trained model artefacts
        """

        return self.aoa_client.get_request(
            path=(f"{self.path}/listObjects").format(trained_model_id),
            header_params=self._get_header_params(),
            query_params={},
        )["objects"]

    def get_signed_download_url(self, trained_model_id: uuid, artefact: str):
        """
        returns a signed url for the artefact

        Parameters:
           trained_model_id (uuid): Trained Model Id
           artefact (str): The artefact to generate the signed url for

        Returns:
            (str): the signed url
        """
        query_params = self.generate_params(["objectKey"], [artefact])

        response = self.aoa_client.get_request(
            path=(f"{self.path}/signedDownloadUrl").format(trained_model_id),
            header_params=self._get_header_params(),
            query_params=query_params,
        )

        return response["endpoint"]

    def get_signed_upload_url(self, trained_model_id: uuid, artefact: str):
        """
        returns a signed url for the artefact

        Parameters:
           trained_model_id (uuid): Trained Model Id
           artefact (str): The artefact to generate the signed url for

        Returns:
            (str): the signed url
        """
        query_params = self.generate_params(["objectKey"], [artefact])

        response = self.aoa_client.get_request(
            path=(f"{self.path}/signedUploadUrl").format(trained_model_id),
            header_params=self._get_header_params(),
            query_params=query_params,
        )

        return response["endpoint"]

    def download_artefacts(self, trained_model_id: uuid, path: str = "."):
        """
        downloads all artefacts for the given trained model

        Parameters:
           trained_model_id (uuid): Trained Model Id
           path (str): the path to download the artefacts to (default cwd)

        Returns:
            None
        """

        for artefact in self.list_artefacts(trained_model_id):
            response = self.aoa_client.session.get(
                self.get_signed_download_url(trained_model_id, artefact)
            )

            output_file = "{}/{}".format(path, artefact)
            if not os.path.exists(os.path.dirname(output_file)):
                os.makedirs(os.path.dirname(output_file))

            with open(output_file, "wb") as f:
                for chunk in response.iter_content(chunk_size=1024 * 1024):
                    f.write(chunk)

    def upload_artefacts(
        self, import_id: uuid, artefacts: List = None, artefacts_folder: str = None
    ):
        """
        uploads artefacts for the given trained model

        Parameters:
           import_id (uuid): Trained Model Id
           artefacts (List): The artefact paths to upload (must specify this or artefacts_folder)
           artefacts_folder (str): The artefact folder (must specify this or artefacts list)
        Returns:
            None
        """

        if artefacts is None and artefacts_folder is None:
            raise ValueError(
                "Either artefacts or artefacts_folder argument must be specified"
            )

        if artefacts is not None:
            for artefact in artefacts:
                object_key = os.path.basename(artefact)
                self.__upload_artefact(artefact, object_key, import_id)

        else:
            for root, d, files in os.walk(artefacts_folder):
                for file in files:
                    object_key = os.path.relpath(
                        os.path.join(root, file), artefacts_folder
                    )
                    self.__upload_artefact(
                        os.path.join(root, file), object_key, import_id
                    )

    def upload_byom_model(self, format: str, file_path: str, import_id: str = None):
        """Simplified method to upload just a single file as BYOM model version.
        File is renamed during upload, so any local filename could be used.

        Args:
            format (str): One of supported formats, e.g. PMML, ONNX, H2O
            file_path (str): local file path to a model artefact
            import_id (str, optional): by default import_id is randomly generated. If required, specific import_id could be provided.

        Raises:
            ValueError: if format string is not one of supported formats

        Returns:
            str: import_id that could be used by import_byom method
        """
        if format not in SUPPORTED_BYOM_FORMATS:
            raise ValueError(
                f"The format {format} is not supported for simplified BYOM import, use"
                " `import_file` instead"
            )

        if not import_id:
            import_id = str(uuid.uuid4())

        self.__upload_artefact(file_path, f"model.{format.lower()}", import_id)
        return import_id

    def __upload_artefact(self, artefact, object_key, import_id):
        signed_url = self.get_signed_upload_url(import_id, object_key)
        # don't use aoa_client.session here as we don't want to send auth info.
        upload_resp = requests.put(
            signed_url, data=open(artefact, "rb"), verify=self.aoa_client.session.verify
        )
        upload_resp.raise_for_status()



import json
import os
import logging
import math
import numbers

from typing import List, Dict
from teradataml.analytics.valib import *
from teradataml import DataFrame
from aoa.context.model_context import ModelContext
from aoa.stats.stats_util import (
    _capture_stats,
    _NpEncoder,
    _parse_scoring_stats,
    _compute_continuous_edges,
)

logger = logging.getLogger(__name__)


def record_training_stats(
    df: DataFrame,
    features: List[str],
    targets: List[str] = [],
    categorical: List[str] = [],
    context: ModelContext = {},
    feature_importance: Dict[str, float] = {},
    **kwargs,
) -> Dict:
    """
    Compute and record the dataset statistics used for training. This information provides ModelOps with a snapshot
    of the dataset at this point in time (i.e. at the point of training). ModelOps uses this information for data and
    prediction drift monitoring. It can also be used for data quality monitoring as all of the information which is
    captured here is available to configure an alert on (e.g. max > some_threshold).

    Depending on the type of variable (categorical or continuous), different statistics and distributions are computed.
    All of this is computed in Vantage via the Vantage Analytics Library (VAL).

    Continuous Variable:
        Distribution: Histogram
        Statistics: Min, Max, Average, Skew, etc, nulls

    Categorical Variable:
        Distribution: Frequency
        Statistics: nulls

    The following example shows how you would use this function for a binary classification problem where the there
    are 3 features and 1 target. As it is classification, the target must be categorical and in this case, the features
    are all continuous.
    example usage:
        training_df = DataFrame.from_query("SELECT * from my_table")

        record_training_stats(training_df,
                              features=["feat1", "feat2", "feat3"],
                              targets=["targ1"],
                              categorical=["targ1"],
                              context=context)

    :param df: teradataml dataframe used for training with the feature and target variables
    :type df: teradataml.DataFrame
    :param features: feature variable(s) used in this training
    :type features: List[str]
    :param targets: target variable(s) used in this training
    :type targets: List[str]
    :param categorical: variable(s) (feature or target) that is categorical
    :type categorical: List[str]
    :param context: ModelContext which is associated with that training invocation
    :type context: ModelContext
    :param feature_importance: (Optional) feature importance
    :type feature_importance: Dict[str, float]
    :return: the computed data statistics
    :rtype: Dict
    :raise ValueError: if features or targets are not provided
    :raise TypeError: if df is not of type teradataml.DataFrame
    """

    logger.info("Computing training dataset statistics")

    if not features:
        raise ValueError("One or more features must be provided")

    # backward compatibility for when we had targets incorrectly named predictors.. remove in future version
    if "predictors" in kwargs:
        logger.warning(
            "Usage of `predictors` as argument to `record_training_stats` is"
            " deprecated. "
        )
        targets = kwargs.pop("predictors")

    if not targets:
        logger.warning(
            "One or more targets are not provided to collect training statistics, make"
            " sure this is what you want."
        )

    feature_metadata_fqtn = None
    feature_metadata_group = None
    data_stats_filename = "artifacts/output/data_stats.json"

    if context:
        feature_metadata_fqtn = context.dataset_info.get_feature_metadata_fqtn()
        feature_metadata_group = context.dataset_info.feature_metadata_monitoring_group
        data_stats_filename = os.path.join(
            context.artifact_output_path, "data_stats.json"
        )

    data_stats = _capture_stats(
        df=df,
        features=features,
        targets=targets,
        categorical=categorical,
        feature_importance=feature_importance,
        feature_metadata_fqtn=feature_metadata_fqtn,
        feature_metadata_group=feature_metadata_group,
    )

    with open(data_stats_filename, "w+") as f:
        json.dump(data_stats, f, indent=2, cls=_NpEncoder)

    return data_stats


def record_evaluation_stats(
    features_df: DataFrame,
    predicted_df: DataFrame,
    feature_importance: Dict[str, float] = {},
    context: ModelContext = None,
    **kwargs,
) -> Dict:
    """
    Compute and record the dataset statistics used for evaluation. This information provides ModelOps with a snapshot
    of the dataset at this point in time (i.e. at the point of evaluation). ModelOps uses this information for data
    and prediction drift monitoring. It can also be used for data quality monitoring as all of the information which
    is captured here is available to configure an alert on (e.g. max > some_threshold).

    Depending on the type of variable (categorical or continuous), different statistics and distributions are computed.
    All of this is computed in Vantage via the Vantage Analytics Library (VAL).

    Continuous Variable:
        Distribution: Histogram
        Statistics: Min, Max, Average, Skew, etc, nulls

    Categorical Variable:
        Distribution: Frequency
        Statistics: nulls

    example usage:
        features_df = DataFrame.from_query("SELECT * from my_features_table")

        predicted_df = model.predict(features_df)

        record_evaluation_stats(features_df=features_df,
                                predicted_df=predicted_df,
                                context=context)

    :param features_df: dataframe containing feature variable(s) from evaluation
    :type features_df: teradataml.DataFrame
    :param predicted_df: dataframe containing predicted target variable(s) from evaluation
    :type predicted_df: teradataml.DataFrame
    :param context: ModelContext which is associated with that training invocation
    :type context: ModelContext
    :param feature_importance: (Optional) feature importance
    :type feature_importance: Dict[str, float]
    :return: the computed data statistics
    :rtype: Dict
    :raise ValueError: if the number of predictions (rows) do not match the number of features (rows)
    :raise TypeError: if features_df or predicted_df is not of type teradataml.DataFrame
    """

    logger.info("Computing evaluation dataset statistics")

    feature_metadata_fqtn = None
    feature_metadata_group = None
    output_data_stats_filename = "artifacts/output/data_stats.json"
    input_data_stats_filename = "artifacts/input/data_stats.json"

    if context:
        feature_metadata_fqtn = context.dataset_info.get_feature_metadata_fqtn()
        feature_metadata_group = context.dataset_info.feature_metadata_monitoring_group
        output_data_stats_filename = os.path.join(
            context.artifact_output_path, "data_stats.json"
        )
        input_data_stats_filename = os.path.join(
            context.artifact_input_path, "data_stats.json"
        )

    with open(input_data_stats_filename, "r") as f:
        training_data_stats = json.load(f)

    data_stats = _parse_scoring_stats(
        features_df=features_df,
        predicted_df=predicted_df,
        data_stats=training_data_stats,
        feature_importance=feature_importance,
        feature_metadata_fqtn=feature_metadata_fqtn,
        feature_metadata_group=feature_metadata_group,
    )

    # for evaluation, the core will do it (we may change this later to unify)..
    with open(output_data_stats_filename, "w+") as f:
        json.dump(data_stats, f, indent=2, cls=_NpEncoder)

    return data_stats


def record_scoring_stats(
    features_df: DataFrame, predicted_df: DataFrame, context: ModelContext = None
) -> Dict:
    """
    Compute and record the dataset statistics used for scoring. This information provides ModelOps with a snapshot
    of the dataset at this point in time (i.e. at the point of scoring). ModelOps uses this information for data
    and prediction drift monitoring. It can also be used for data quality monitoring as all of the information which
    is captured here is available to configure an alert on (e.g. max > some_threshold).

    Depending on the type of variable (categorical or continuous), different statistics and distributions are computed.
    All of this is computed in Vantage via the Vantage Analytics Library (VAL).

    Continuous Variable:
        Distribution: Histogram
        Statistics: Min, Max, Average, Skew, etc, nulls

    Categorical Variable:
        Distribution: Frequency
        Statistics: nulls

    example usage:
        features_df = DataFrame.from_query("SELECT * from my_features_table")

        predicted_df = model.predict(features_df)

        record_scoring_stats(features_df=features_df,
                            predicted_df=predicted_df,
                            context=context)

    :param features_df: dataframe containing feature variable(s) from evaluation
    :type features_df: teradataml.DataFrame
    :param predicted_df: dataframe containing predicted target variable(s) from evaluation
    :type predicted_df: teradataml.DataFrame
    :param context: ModelContext which is associated with that training invocation
    :type context: ModelContext
    :return: the computed data statistics
    :rtype: Dict
    :raise ValueError: if the number of predictions (rows) do not match the number of features (rows)
    :raise TypeError: if features_df or predicted_df is not of type teradataml.DataFrame
    """

    logger.info("Computing scoring dataset statistics")

    feature_metadata_fqtn = None
    feature_metadata_group = None
    input_data_stats_filename = "artifacts/input/data_stats.json"

    if context:
        feature_metadata_fqtn = context.dataset_info.get_feature_metadata_fqtn()
        feature_metadata_group = context.dataset_info.feature_metadata_monitoring_group
        input_data_stats_filename = os.path.join(
            context.artifact_input_path, "data_stats.json"
        )
        output_data_stats_filename = os.path.join(
            context.artifact_output_path, "data_stats.json"
        )

    with open(input_data_stats_filename, "r") as f:
        training_data_stats = json.load(f)

    data_stats = _parse_scoring_stats(
        features_df=features_df,
        predicted_df=predicted_df,
        data_stats=training_data_stats,
        feature_metadata_fqtn=feature_metadata_fqtn,
        feature_metadata_group=feature_metadata_group,
    )

    # for evaluation, the core will do it (we may change this later to unify)..
    with open(output_data_stats_filename, "w+") as f:
        json.dump(data_stats, f, indent=2, cls=_NpEncoder)

    return data_stats


def compute_continuous_stats(features_df: DataFrame, continuous_features: List[str]):
    """This function computes bin edges for continuous features. Only numeric columns are used,
    others are ignored with warning. For each column it computes maximum and minimum values, and
    attempts to split the difference into 10 bins. If only smaller number is possile - it generates
    the maximum number (e.g. integer column with minimum 5 and maximum 10 generates only 5 bins).
    The column has to have at least two distinct values, otherwise the column is ignored.

    Args:
        features_df (DataFrame): Teradata DataFrame used to compute statistics metadata
        continuous_features (List): list of columns representing continuous features

    Returns:
        dict: Dictionary with keys corresponding to requested features, and values containing edges
    """
    dtypes = {r[0].lower(): r[1] for r in features_df.dtypes._column_names_and_types}

    # Numeric types in teradata are represented as either int, or float or decimal.Decimal,
    # so other columns shall be ignored
    # COLUMN NAME	TYPE
    # int_column	int
    # byte_column	int
    # smallint_column	int
    # decimal_column	decimal.Decimal
    # numeric_column	decimal.Decimal
    # float_column	float
    # real_column	float
    # double_precision_column	float
    # number_column	decimal.Decimal
    lowered_features = list(map(str.lower, continuous_features))
    features = []
    for f in lowered_features:
        if (
            dtypes[f].startswith("decimal")
            or dtypes[f].startswith("float")
            or dtypes[f].startswith("int")
        ):
            features.append(f)
        else:
            logger.warning(
                f"Column {f} has a type {dtypes[f]} which is not compatible with"
                " continuous feature types. This column will be ignored."
            )

    logger.debug(
        "Executing this VAL"
        f" {valib.Statistics(data=features_df, columns=features, stats_options='all').show_query()}"
    )
    stats = valib.Statistics(data=features_df, columns=features, stats_options="all")
    stats = stats.result.to_pandas(all_rows=True).reset_index()
    stats["xcol"] = stats["xcol"].str.lower()
    stats = stats.set_index("xcol", drop=False)

    non_trivial_features = []
    for f in features:
        if stats.loc[f].loc["xcnt"] == 0:
            logger.warning(f"Feature {f} has only NULL values, ignored")
        elif stats.loc[f].loc["xmin"] == stats.loc[f].loc["xmax"]:
            logger.warning(
                f"Feature {f} doesn't have enough unique values to be considered"
                " continuous feature (needs at least two distinct not null values),"
                " ignored"
            )
        else:
            non_trivial_features.append(f)

    bins = 10  # Default number of bins; if data doesn't allow 10 bins - we generate the maximum number allowed
    reference_edges = _compute_continuous_edges(
        non_trivial_features, stats, dtypes, bins=bins
    )
    edges_dict = dict(zip(non_trivial_features, reference_edges))

    for variable_name, edges in edges_dict.items():
        if len(edges) < bins:
            logger.warning(
                f"Variable {variable_name} has only {len(edges)} bins computed when"
                f" {bins} should have been computed {edges}.\nPlease ensure the"
                " variable is not categorical (use -t categorical)."
            )

    column_stats = {}
    for f in edges_dict.keys():
        column_stats[f.lower()] = {"edges": edges_dict[f]}

    if len(column_stats) == 0:
        raise Exception(
            f"No columns with computable statistics metadata were found, please see"
            f" warning messages above"
        )

    return column_stats


def compute_categorical_stats(features_df: DataFrame, categorical_features: List[str]):
    """This function computes frequencies for categorical features. Each column must have at least one non-NULL value,
    all NULL columns are ignored. NULL value is ignored in frequencies (number of NULLs is reported for every
    training/evaluation/scoring jobs).

    Args:
        features_df (DataFrame): Teradata DataFrame used to compute statistics metadata
        categorical_features (List): list of columns representing categorical features

    Returns:
        dict: Dictionary with keys corresponding to requested features, and values containing frequencies
    """
    logger.debug(
        "Executing this VAL"
        f" {valib.Frequency(data=features_df, columns=categorical_features).show_query()}"
    )
    statistics = valib.Frequency(data=features_df, columns=categorical_features)
    statistics = statistics.result.to_pandas(all_rows=True).reset_index()
    statistics = statistics.drop(
        statistics.columns.difference(["xcol", "xval", "xpct"]), axis=1
    )
    statistics["xcol"] = statistics["xcol"].str.lower()
    statistics = (
        statistics.groupby("xcol")
        .apply(lambda x: dict(zip(x["xval"], x["xpct"])))
        .to_dict()
    )

    lowered_features = list(map(str.lower, categorical_features))
    features = list(lowered_features)
    for f in features:
        values_list = list(statistics[f].keys())
        for k in values_list:
            if isinstance(k, numbers.Number):
                if math.isnan(k):
                    logger.warning(
                        f"Categorical feature {f} has NULL values in reference table,"
                        " NULLs will be ignored"
                    )
                    del statistics[f][k]
        if not statistics[f]:
            logger.warning(
                f"Categorical feature {f} has only NULL values in reference table, no"
                " statistics metadata generated"
            )
            lowered_features.remove(f)

    column_stats = {
        f: {"categories": list(statistics[f].keys())} for f in lowered_features
    }

    return column_stats



import runpy
import types
import inspect
from aoa.context.model_context import ModelContext, DatasetInfo
from aoa.cli.base_task import BaseTask
import shutil
import os
import logging


class RunTask(BaseTask):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def run_task_local(self, base_path: str, func: str = None):

        base_path = os.path.join(base_path, "")
        task_path = base_path + "feature_engineering/"

        if not os.path.exists(task_path):
            raise ValueError(
                "feature engineering task directory {0} does not exist".format(
                    task_path
                )
            )

        task_artefacts_path = ".artefacts/FE/"
        task_artefacts_output_path = ".artefacts/FE/output/"
        if os.path.exists(task_artefacts_path):
            self.logger.debug(
                "Cleaning local model artefact path: {}".format(task_artefacts_path)
            )
            shutil.rmtree(task_artefacts_path)

        os.makedirs(task_artefacts_output_path)

        try:
            if os.path.exists("./artifacts"):
                os.remove("./artifacts")

            os.symlink(task_artefacts_path, "./artifacts", target_is_directory=True)

            cli_task_kargs = self._BaseTask__get_task_varargs()

            context = ModelContext(
                artifact_output_path="artifacts/output",
                dataset_info=DatasetInfo.from_dict({}),
                hyperparams={},
                **cli_task_kargs,
            )

            task_file_path = os.path.join(task_path, "task.py")
            if not os.path.exists(task_file_path):
                raise ValueError(f"Task file {task_file_path} does not exist")

            task_module = runpy.run_path(task_file_path)

            if not func:
                function_names = [
                    name
                    for name, value in task_module.items()
                    if isinstance(value, types.FunctionType)
                    and inspect.getfile(value) == task_file_path
                ]
                func = self.__input_select(
                    "function", function_names, "Available functions:"
                )

            print("")

            if func in task_module and isinstance(
                task_module[func], types.FunctionType
            ):
                self.logger.info("Loading and executing task code")
                task_module[func](context=context)
            else:
                raise ValueError(f"{func} function not found in task.py")

            self.logger.info(
                "Artefacts can be found in: {}".format(task_artefacts_output_path)
            )
            self.__cleanup()
            return func
        except ModuleNotFoundError:
            self.__cleanup()
            self.logger.error(
                "Missing required python module, try running following command first:"
            )
            self.logger.error("pip install -r {}requirements.txt".format(task_path))
            raise
        except:
            self.__cleanup()
            self.logger.exception("Exception running feature engineering task code")
            raise

    def __cleanup(self):
        if os.path.exists("./artifacts"):
            os.remove("./artifacts")

    def __print_underscored(self, message):
        print(message)
        print("-" * len(message))

    def __input_select(self, name, values, label="", default=None):
        if len(values) == 0:
            return

        if label != "":
            self.__print_underscored(label)

        for ix, item in enumerate(values):
            default_text = " (default)" if default and default == item else ""
            print("[{}] {}".format(ix, item) + default_text)

        tmp_index = (
            input(
                "Select {} by index (or leave blank for the default one): ".format(name)
            )
            if default
            else input("Select {} by index: ".format(name))
        )

        if default and default in values and tmp_index == "":
            tmp_index = values.index(default)
        elif (tmp_index == "" and not default) or (
            not tmp_index.isnumeric() or int(tmp_index) >= len(values)
        ):
            print(
                "Wrong selection, please try again by selecting the index number on the"
                " first column."
            )
            print("You may cancel at anytime by pressing Ctrl+C.")
            print("")
            return self.__input_select(name, values, label, default)

        return values[int(tmp_index)]



from __future__ import absolute_import
from typing import Dict, List

from aoa.api.iterator_base_api import IteratorBaseApi


class DatasetTemplateApi(IteratorBaseApi):
    path = "/api/datasetTemplates"
    type = "DATASET_TEMPLATE"

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.aoa_client.select_header_accept([
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]),
        ]

        return self.generate_params(header_vars, header_vals)

    def find_by_name_like(self, name: str, projection: str = None) -> List:
        """
        returns a list dataset template of a project by name

        Parameters:
           name (str): dataset name(string) to find
           projection (str): projection type

        Returns:
            (list): dataset template
        """
        query_vars = ["name", "projection"]
        query_vals = [name, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByName",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def save(self, dataset_template: Dict[str, str]):
        """
        register a dataset template

        Parameters:
           dataset template (dict): dataset template to register

        Returns:
            (dict): dataset template
        """
        return self.aoa_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=dataset_template,
        )

    def render(self, id: str) -> Dict:
        """
        returns a rendered dataset template

        Parameters:
           id (str): dataset_template id

        Returns:
            (dict): rendered dataset template
        """
        return self.aoa_client.get_request(
            path=f"{self.path}/{id}/render",
            header_params=self._get_header_params(),
            query_params={},
        )



from __future__ import absolute_import

import aia
import json
import logging
import oauthlib
import os
import re
import requests
import time
import yaml

from oauthlib.oauth2 import BackendApplicationClient
from typing import List, Dict
from requests_oauthlib import OAuth2Session
from requests.auth import HTTPBasicAuth
from .spinner import spin_it

os.system("")  # enables ansi escape characters in terminal


class AoaClient(object):
    DEFAULT_TOKEN_CACHE_FILE_PATH = os.path.join(os.path.expanduser("~/.aoa"), ".token")
    DEFAULT_CONFIG_FILE_PATH = os.path.join(os.path.expanduser("~/.aoa"), "config.yaml")

    def __init__(self, **kwargs):
        """

        Parameters:
           project_id (str): project id(uuid)
           ssl_verify (bool): enable or disable TLS cert validation
           aoa_url (str): aoa api endpoint

           auth_mode (str): device_code|client_credentials|bearer

           If using auth_mode=bearer
           auth_bearer (str): the raw bearer token

           If using auth_mode=client_credentials
           auth_client_id (str): oauth2 client id
           auth_client_secret (str): oauth2 client secret
           auth_token_url (str): oauth2 token endpoint

           if using auth_mode=device_code
           auth_client_id (str): oauth2 client id
           auth_client_secret (str): oauth2 client secret
           auth_token_url (str): oauth2 token endpoint
           auth_device_auth_url (str): oauth2 device code endpoint
        """
        self.logger = logging.getLogger(__name__)

        self.project_id = kwargs.get("project_id")
        self.ssl_verify = True
        self.aoa_url = None

        self.auth_mode = None

        self.__client_id = None
        self.__client_secret = None
        self.__token_url = None
        self.__device_auth_url = None
        self.__bearer_token = None

        self.__parse_aoa_config(**kwargs)

        if self.auth_mode == "device_code":
            try:
                self.__create_oauth_session_device_code()
            except oauthlib.oauth2.rfc6749.errors.InvalidGrantError as ge:
                if ge.description in ["Token is not active", "Session not active"]:
                    logging.warning(ge.description + "\nRetrying one more time\n")

                    self.__remove_cached_token()

                    self.__create_oauth_session_device_code()
                else:
                    raise ge
            except oauthlib.oauth2.rfc6749.errors.InvalidTokenError as ge:
                logging.warning(ge.description + "\nRetrying one more time\n")

                self.__remove_cached_token()

                self.__create_oauth_session_device_code()

        elif self.auth_mode == "client_credentials":
            self.__create_oauth_session_client_credentials()

        elif self.auth_mode == "bearer":
            self.__create_oauth_session_bearer()

        elif self.auth_mode is None:
            raise Exception(
                "ModelOps endpoint not configured.\nTry (re)copy the CLI configuration"
                " from ModelOps UI -> Session Details -> CLI Config."
            )

        else:
            raise Exception(f"Auth mode: {self.auth_mode} not supported.")

    def __parse_aoa_config(self, **kwargs):
        if "config_file" in kwargs:
            self.__parse_yaml(kwargs["config_file"])
        else:
            if os.path.isfile(self.DEFAULT_CONFIG_FILE_PATH):
                self.__parse_yaml(self.DEFAULT_CONFIG_FILE_PATH)

        self.__parse_env_variables()
        self.__parse_kwargs(**kwargs)

    def __parse_yaml(self, yaml_path: str):
        with open(yaml_path, "r") as handle:
            conf = yaml.safe_load(handle)
        self.__parse_kwargs(**conf)

    def __parse_kwargs(self, **kwargs):
        self.aoa_url = kwargs.get("aoa_url", self.aoa_url)
        self.ssl_verify = kwargs.get("ssl_verify", self.ssl_verify)
        self.auth_mode = kwargs.get("auth_mode", self.auth_mode)

        self.__check_legacy_mode()

        if self.auth_mode == "device_code":
            self.__client_id = kwargs.get("auth_client_id", self.__client_id)
            self.__client_secret = kwargs.get(
                "auth_client_secret", self.__client_secret
            )
            self.__token_url = kwargs.get("auth_token_url", self.__token_url)
            self.__device_auth_url = kwargs.get(
                "auth_device_auth_url", self.__device_auth_url
            )

        elif self.auth_mode == "client_credentials":
            self.__client_id = kwargs.get("auth_client_id", self.__client_id)
            self.__client_secret = kwargs.get(
                "auth_client_secret", self.__client_secret
            )
            self.__token_url = kwargs.get("auth_token_url", self.__token_url)

        elif self.auth_mode == "bearer":
            self.__bearer_token = kwargs.get("auth_bearer", self.__bearer_token)

        if "verify_connection" in kwargs:
            self.verify_aoa_connection = kwargs["verify_connection"]

    def __parse_env_variables(self):
        self.aoa_url = os.environ.get("AOA_URL", self.aoa_url)
        self.ssl_verify = (
            os.environ.get("AOA_SSL_VERIFY", str(self.ssl_verify)).lower() == "true"
        )
        self.auth_mode = os.environ.get("AOA_API_AUTH_MODE", self.auth_mode)

        self.__check_legacy_mode()

        if self.auth_mode == "device_code":
            self.__client_id = os.environ.get(
                "AOA_API_AUTH_CLIENT_ID", self.__client_id
            )
            self.__client_secret = os.environ.get(
                "AOA_API_AUTH_CLIENT_SECRET", self.__client_secret
            )
            self.__token_url = os.environ.get(
                "AOA_API_AUTH_TOKEN_URL", self.__token_url
            )
            self.__device_auth_url = os.environ.get(
                "AOA_API_AUTH_DEVICE_AUTH_URL", self.__device_auth_url
            )

        elif self.auth_mode == "client_credentials":
            self.__client_id = os.environ.get(
                "AOA_API_AUTH_CLIENT_ID", self.__client_id
            )
            self.__client_secret = os.environ.get(
                "AOA_API_AUTH_CLIENT_SECRET", self.__client_secret
            )
            self.__token_url = os.environ.get(
                "AOA_API_AUTH_TOKEN_URL", self.__token_url
            )

        elif self.auth_mode == "bearer":
            self.__bearer_token = os.environ.get(
                "AOA_API_AUTH_BEARER_TOKEN", self.__bearer_token
            )

    def __check_legacy_mode(self):
        if self.auth_mode == "oauth-cc":
            self.auth_mode = "client_credentials"
        elif self.auth_mode == "oauth":
            self.auth_mode = "device_code"

    def __validate_url(self):
        if not self.aoa_url:
            raise ValueError(
                "ModelOps endpoint not configured.\nTry (re)copy the CLI configuration"
                " from ModelOps UI -> Session Details -> CLI Config."
            )

    def set_project_id(self, project_id: str):
        """
        set project id

        Parameters:
           project_id (str): project id(uuid)
        """
        self.project_id = project_id

    def get_current_project(self):
        """
        get project id

        Return:
           project_id (str): project id(uuid)
        """
        return self.project_id

    @staticmethod
    def select_header_accept(accepts: List[str]):
        """
        converts list of header into a string

        Return:
            (str): request header
        """
        if not accepts:
            return

        accepts = [x.lower() for x in accepts]
        return ", ".join(accepts)

    def get_request(
        self, path, header_params: Dict[str, str], query_params: Dict[str, str]
    ):
        """
        wrapper for get request

        Parameters:
           path (str): url
           header_params (dict): header parameters
           query_params (dict): query parameters

        Returns:
            dict for resources, str for errors, None for 404
        Raise:
            raises HTTPError in case of error status code other than 404
        """

        self.__validate_url()

        resp = self.session.get(
            url=self.__strip_url(self.aoa_url) + path,
            headers=header_params,
            params=query_params,
        )

        if resp.status_code == 404:
            return None

        return self.__validate_and_extract_body(resp)

    def post_request(
        self,
        path,
        header_params: Dict[str, str],
        query_params: Dict[str, str],
        body: Dict[str, str],
    ):
        """
        wrapper for post request

        Parameters:
           path (str): url
           header_params (dict): header parameters
           query_params (dict): query parameters
           body (dict): request body

        Returns:
            dict for resources, str for errors
        Raise:
            raises HTTPError in case of error status code
        """

        self.__validate_url()

        resp = self.session.post(
            url=self.__strip_url(self.aoa_url) + path,
            headers=header_params,
            params=query_params,
            data=json.dumps(body),
        )

        return self.__validate_and_extract_body(resp)

    def put_request(
        self,
        path,
        header_params: Dict[str, str],
        query_params: Dict[str, str],
        body: Dict[str, str],
    ):
        """
        wrapper for put request

        Parameters:
           path (str): url
           header_params (dict): header parameters
           query_params (dict): query parameters
           body (dict): request body

        Returns:
            dict for resources, str for errors
        Raise:
            raises HTTPError in case of error status code
        """

        self.__validate_url()

        resp = self.session.put(
            url=self.__strip_url(self.aoa_url) + path,
            headers=header_params,
            params=query_params,
            data=json.dumps(body),
        )

        return self.__validate_and_extract_body(resp)

    def delete_request(
        self,
        path,
        header_params: Dict[str, str],
        query_params: Dict[str, str],
        body: Dict[str, str],
    ):
        """
        wrapper for delete request
        Parameters:
           path (str): url
           header_params (dict): header parameters
           query_params (dict): query parameters
           body (dict): request body
        Returns:
            dict for resources, str for errors
        Raise:
            raises HTTPError in case of error status code
        """

        self.__validate_url()

        resp = self.session.delete(
            url=self.__strip_url(self.aoa_url) + path,
            headers=header_params,
            params=query_params,
            data=json.dumps(body),
        )

        return self.__validate_and_extract_body(resp)

    def __validate_and_extract_body(self, resp):
        if resp.status_code == 401:
            self.__remove_cached_token()
            self.logger.warning(
                "Clearing the token cache. Please re-run cmd and login again."
            )

        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as err:
            if resp.text:
                raise requests.exceptions.HTTPError(f"Error message: {resp.text}")
            else:
                raise err

        try:
            return resp.json()
        except ValueError:
            return resp.text

    @staticmethod
    def __strip_url(url):
        return url.rstrip("/")

    def __create_oauth_session_device_code(self):
        self.logger.debug("Configuring oauth with device_code grant")

        if (
            self.__client_id is None
            or self.__token_url is None
            or self.__device_auth_url is None
        ):
            raise Exception(
                "Missing CLI configuration.\n"
                + "Please (re)copy the CLI configuration from "
                + "ModelOps UI -> Session Details -> CLI Config\n"
            )

        token = None
        if os.path.exists(self.DEFAULT_TOKEN_CACHE_FILE_PATH):
            self.logger.debug(
                f"Loading cached token data from {self.DEFAULT_TOKEN_CACHE_FILE_PATH}"
            )
            with open(self.DEFAULT_TOKEN_CACHE_FILE_PATH, "r") as f:
                token = json.load(f)

        if not token:
            self.session = requests.session()
            self.__set_session_tls()
            token = self.__get_device_code()

        if "access_token" in token:
            self.logger.debug(f"Access token acquired successfully: {token}")

            session = OAuth2Session(client_id=self.__client_id)

            # don't chase certs/print warning for TLS if already done for __get_device_code
            if hasattr(self, "session"):
                session.verify = self.session.verify
                self.session = session
            else:
                self.session = session
                self.__set_session_tls()

            self.__bearer_token = f"Bearer {token['access_token']}"
            self.__create_oauth_session_bearer()

            with open(self.DEFAULT_TOKEN_CACHE_FILE_PATH, "w") as f:
                json.dump(token, f)

        elif "refresh_token" in token:
            self.logger.debug(f"Refresh token acquired successfully: {token}")

            session = OAuth2Session(client_id=self.__client_id)

            # don't chase certs/print warning for TLS if already done for __get_device_code
            if hasattr(self, "session"):
                session.verify = self.session.verify
                self.session = session
            else:
                self.session = session
                self.__set_session_tls()

            self.session.refresh_token(
                token_url=self.__token_url,
                refresh_token=token["refresh_token"],
                auth=HTTPBasicAuth(self.__client_id, self.__client_secret),
                verify=self.ssl_verify,
            )

            with open(self.DEFAULT_TOKEN_CACHE_FILE_PATH, "w") as f:
                json.dump(token, f)

        else:
            raise Exception(
                "Token does not contain access_token or refresh_token. Received"
                f" {token}"
            )

    def __create_oauth_session_client_credentials(self):
        self.logger.debug("Configuring oauth with client_credentials grant")

        if (
            self.__client_id is None
            or self.__client_secret is None
            or self.__token_url is None
        ):
            raise Exception(
                "AOA_API_AUTH_CLIENT_ID, AOA_API_AUTH_CLIENT_SECRET, "
                "AOA_API_AUTH_TOKEN_URL must be defined "
                "with AOA_API_AUTH_MODE of 'client_credentials'"
            )

        self.session = OAuth2Session(
            client=BackendApplicationClient(client_id=self.__client_id)
        )
        self.__set_session_tls()
        self.session.fetch_token(
            token_url=self.__token_url,
            auth=HTTPBasicAuth(self.__client_id, self.__client_secret),
            verify=self.ssl_verify,
        )

    def __create_oauth_session_bearer(self):
        self.session = requests.session()
        self.session.headers.update({"Authorization": self.__bearer_token})
        self.__set_session_tls()

    def __get_device_code(self):
        device_code_response = self.session.post(
            self.__device_auth_url,
            data={"client_id": self.__client_id, "scope": "openid profile"},
        )

        if device_code_response.status_code != 200:
            raise Exception(
                "Error generating the device code. Received code"
                f" {device_code_response.status_code}."
            )

        device_code_data = device_code_response.json()
        print(
            "1. On your computer or mobile device navigate to: ",
            device_code_data["verification_uri_complete"],
        )
        print("2. Enter the following code: ", device_code_data["user_code"])

        def authorize():
            authenticated = False
            token_data = None

            while not authenticated:
                token_response = self.session.post(
                    self.__token_url,
                    data={
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                        "device_code": device_code_data["device_code"],
                        "client_id": self.__client_id,
                    },
                )

                token_data = token_response.json()
                if token_response.status_code == 200:
                    authenticated = True

                elif "error" in token_data:
                    if token_data["error"] in ("authorization_pending", "slow_down"):
                        time.sleep(device_code_data["interval"])
                    else:
                        raise Exception(token_data["error_description"])

                else:
                    raise Exception(f"Bad response code {token_response.status_code}")

            return token_data

        msg = "Waiting for device code to be authorized"
        res = spin_it(authorize, msg, 0.2)
        print(
            "\033[32m\U0001f512 This device has been authorized successfully.\033[0m\n"
        )
        return res

    def __set_session_tls(self):
        self.session.verify = self.ssl_verify
        if self.ssl_verify:
            self.__chase_tls_cert_chain()
        else:
            from requests.packages.urllib3.exceptions import InsecureRequestWarning

            self.logger.warning(
                "Certificate validation disabled. Adding certificate verification is"
                " strongly advised"
            )
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

    def __chase_tls_cert_chain(self):
        if self.aoa_url.startswith("https") and not (
            "REQUESTS_CA_BUNDLE" in os.environ or "CURL_CA_BUNDLE" in os.environ
        ):

            try:
                if re.search("http|https", self.aoa_url):
                    requests.get(f"{self.aoa_url}/admin/info")
                else:
                    raise Exception("Invalid AOA_URL")
            except requests.exceptions.SSLError:
                from aia import AIASession
                from tempfile import NamedTemporaryFile

                self.logger.debug("Attempting certificate chain chasing via aia")

                # unless, debug logging enabled,
                # change logging level for aia to warning as it prints debug at info level
                if logging.root.level > logging.DEBUG:
                    logging.getLogger("aia").setLevel(logging.WARNING)

                try:
                    aia_session = AIASession()

                    ca_data = aia_session.cadata_from_url(self.aoa_url)
                    with NamedTemporaryFile("w", delete=False) as pem_file:
                        pem_file.write(ca_data)
                        pem_file.flush()

                    self.session.verify = pem_file.name
                except aia.InvalidCAError:
                    raise Exception(
                        "Attempted to find trusted root certificate via AIA chasing but"
                        " not found.\nPlease configure REQUESTS_CA_BUNDLE or"
                        " CURL_CA_BUNDLE.\nAlternatively, to ignore TLS validation (not"
                        " advised), export AOA_SSL_VERIFY=false"
                    )

    @staticmethod
    def __remove_cached_token():
        if os.path.exists(AoaClient.DEFAULT_TOKEN_CACHE_FILE_PATH):
            os.remove(AoaClient.DEFAULT_TOKEN_CACHE_FILE_PATH)



from __future__ import absolute_import
from typing import List, Dict
from aoa.api_client import AoaClient

import abc


class BaseApi(object):
    path = ""
    json_type = "application/json"

    def __init__(self, aoa_client: AoaClient):
        self.aoa_client = aoa_client

    @abc.abstractmethod
    def _get_header_params(self):
        pass

    def generate_params(self, params: List[str], values: List[str]):
        """
        returns list of parameters and values as dictionary

        Parameters:
           params (List[str]): list of parameter names
           values (List[str]): list of parameter values

        Returns:
            (dict): generated parameters
        """

        # bools in python start with upper case when converted to strs. APIs expect lowercase
        api_values = [str(v).lower() if type(v) is bool else v for v in values]

        return dict(zip(params, api_values))

    def required_params(self, param_names: List[str], dict_obj: Dict[str, str]):
        """
        checks required parameters, raises exception if the required parameter is missing in the dictionary

        Parameters:
           param_names (List[str]): list of required parameter names
           dict_obj (Dict[str, str]): dictionary to check for required parameters
        """
        for param in param_names:
            if param not in dict_obj:
                raise ValueError("Missing required value " + str(param))

    def find_all(
        self,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns all entities

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): all entities
        """
        header_params = self._get_header_params()

        query_vars = ["projection", "page", "size", "sort"]
        query_vals = [projection, page, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(self.path, header_params, query_params)

    def find_by_archived(
        self,
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns all entities by archived

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order e.g. name?asc / name?desc
           archived (bool): whether to return archived or unarchived entities

        Returns:
            (dict): all entities
        """
        header_params = self._get_header_params()

        query_vars = ["projection", "page", "size", "sort", "archived"]
        query_vals = [projection, page, size, sort, archived]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            f"{self.path}/search/findByArchived", header_params, query_params
        )

    def find_by_id(self, id: str, projection: str = None):
        """
        returns the entity

        Parameters:
           id (str): entity id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            f"{self.path}/{id}", header_params, query_params
        )

    def archive(self, id: str):
        """
        archives the entity
        Parameters:
           id (str): entity id(uuid) to archive
        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        return self.aoa_client.post_request(
            f"/api/archives/{self.type}/{id}", header_params, {}, {}
        )

    def unarchive(self, id: str):
        """
        unarchives the entity
        Parameters:
           id (str): entity id(uuid) to unarchive
        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        return self.aoa_client.delete_request(
            f"/api/archives/{self.type}/{id}", header_params, {}, {}
        )



from aoa.api.iterator_base_api import BaseApi


class UserAttributesApi(BaseApi):

    path = "/api/userAttributes"
    type = "USER_ATTRIBUTES"

    def _get_header_params(self):
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([self.json_type, "text/plain", "*/*"]),
        ]

        return self.generate_params(header_vars, header_vals)

    def get_default_connection(self):

        return self.aoa_client.get_request(
            path=f"{self.path}/search/findByName",
            header_params=self._get_header_params(),
            query_params=self.generate_params(["name"], ["DEFAULT_CONNECTION"]),
        )



#!/usr/bin/env python

import argparse
import logging
import oauthlib
import os
import re
import requests
import sys
import traceback
import yaml

from pathlib import Path
from aoa.crypto import crypto

os.system("")  # enables ansi escape characters in terminal

COLOR = {
    "WARN": "\033[91m",
    "END": "\033[0m",
}

base_path = os.path.abspath(os.getcwd())
model_catalog = os.path.join(base_path, "model_definitions/")
available_modes = ["Train", "Evaluate", "Score"]
config_dir = f"{Path.home()}/.aoa"


def validate_cwd_valid():
    return os.path.exists(model_catalog)


def init_model_directory(args, repo_manager, aoa_client, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    repo_manager.init_model_directory()

    if not repo_manager.repo_config_exists():
        link_repo(repo_manager=repo_manager, aoa_client=aoa_client)


def input_string(
    name, required=False, tooltip="", password=False, called_from_test=False
):
    from getpass import getpass

    if tooltip != "":
        print(tooltip)

    value = (
        getpass("Enter {}: ".format(name))
        if password and not called_from_test
        else input("Enter {}: ".format(name))
    )

    if value == "" and required:
        print("Value required. Please try again.")
        print("You may cancel at anytime by pressing Ctrl+C")
        print("")
        return input_string(name, required, tooltip)

    return value


def input_select(name, values, label="", default=None):
    if len(values) == 0:
        return

    if label != "":
        print_underscored(label)

    for ix, item in enumerate(values):
        default_text = " (default)" if default and default == item else ""
        print("[{}] {}".format(ix, item) + default_text)

    tmp_index = (
        input("Select {} by index (or leave blank for the default one): ".format(name))
        if default
        else input("Select {} by index: ".format(name))
    )

    if default and default in values and tmp_index == "":
        tmp_index = values.index(default)
    elif (tmp_index == "" and not default) or (
        not tmp_index.isnumeric() or int(tmp_index) >= len(values)
    ):
        print(
            "Wrong selection, please try again by selecting the index number on the"
            " first column."
        )
        print("You may cancel at anytime by pressing Ctrl+C.")
        print("")
        return input_select(name, values, label, default)

    return values[int(tmp_index)]


def yes_or_no(question):
    while "the answer is invalid":
        reply = str(input(question + " (y/n): ")).lower().strip()
        if reply[:1] == "y":
            return True
        if reply[:1] == "n":
            return False


def bash_escape(string):
    return string.replace("\\", "\\\\")


def add_model(args, repo_manager, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    if not validate_cwd_valid():
        logging.error(
            "Current working directory must the root of a model repository to execute"
            " this command"
        )
        exit(1)

    import tempfile

    temp_path = tempfile.TemporaryDirectory()
    logging.info("Using this model catalog from Git repo {}".format(args.template_url))
    repo_manager.clone_repository(args.template_url, temp_path.name, args.branch)

    templates = repo_manager.get_templates(source_path=temp_path.name)
    if len(templates) == 0:
        logging.error(
            "Repo doesn't contain any models to use as templates, use different repo as"
            " a model catalog"
        )
        exit(1)

    print("")
    model_lang = input_select(
        "model language", list(templates.keys()), "Available languages:"
    )

    print("")
    models = {f"{v[0]} ({k})": k for k, v in templates[model_lang].items()}
    model_template = input_select(
        "template type", list(models.keys()), "Available models:"
    )
    model_template_dir = templates[model_lang][models[model_template]][1]

    model_name = input_string("model name", True)
    model_desc = input_string("model description", False)

    import uuid

    model_id = str(uuid.uuid4())

    repo_manager.add_model(
        model_id=model_id,
        model_name=model_name,
        model_desc=model_desc,
        template=model_template_dir,
        base_path=base_path,
    )
    print("Creating model structure for model: ({}) {}".format(model_id, model_name))


def run_model(args, repo_manager, aoa_client, **kwargs):
    from aoa import TrainModel
    from aoa import EvaluateModel
    from aoa import ScoreModel
    from aoa import DatasetApi
    from aoa import DatasetTemplateApi
    import json

    if args.cwd:
        set_cwd(args.cwd)

    current_project = get_current_project(repo_manager, aoa_client, True)
    if not current_project:
        logging.error(
            "Current working directory must the root of a model repository to execute"
            " this command"
        )
        exit(1)

    aoa_client.set_project_id(current_project["id"])

    def _select_model_id(catalog):
        catalog_values = [catalog[i]["name"] for i in catalog]
        selected_model_value = input_select(
            "model", catalog_values, "Available models:"
        )
        selected_model = next(
            (catalog[i] for i in catalog if catalog[i]["name"] == selected_model_value),
            None,
        )
        print("")
        return selected_model["id"] if "id" in selected_model else None

    def _select_mode():
        selected_mode = input_select(
            "mode", available_modes, "Available modes:"
        ).lower()
        print("")
        return selected_mode

    def _select_dataset(dataset_list):
        dataset_values = [i["name"] for i in dataset_list]
        selected_dataset_value = input_select(
            "dataset", dataset_values, "Available datasets:"
        )
        selected_dataset = next(
            (i for i in dataset_list if i["name"] == selected_dataset_value), None
        )
        print("")
        return selected_dataset

    def _select_dataset_template(template_list):
        template_values = [i["name"] for i in template_list]
        selected_template_value = input_select(
            "dataset template", template_values, "Available dataset templates:"
        )
        selected_template = next(
            (i for i in template_list if i["name"] == selected_template_value), None
        )
        print("")
        return selected_template

    def _select_connection(connection=None):
        from argparse import Namespace

        return activate_connection(Namespace(cwd=None, connection=connection))

    available_models = TrainModel.get_model_ids(model_catalog, True)
    dataset_id = None
    dataset_template_id = None

    if not args.model_id:
        model_id = _select_model_id(available_models)
    else:
        model_id_exists = next(
            (
                available_models[i]
                for i in available_models
                if available_models[i]["id"] == args.model_id
            ),
            False,
        )
        if not model_id_exists:
            print(
                "Model not found. Please select one from the list below or press Ctrl+C"
                " to quit."
            )
        model_id = (
            args.model_id if model_id_exists else _select_model_id(available_models)
        )

    if not args.mode:
        mode = _select_mode()

    else:
        mode_exists = (
            True if args.mode in [x.lower() for x in available_modes] else False
        )
        if not mode_exists:
            print(
                "Mode not found. Please select one from the list below or press Ctrl+C"
                " to quit."
            )
        mode = args.mode if mode_exists else _select_mode()

    if args.local_dataset:
        with open(args.local_dataset, "r") as f:
            rendered_dataset = json.load(f)

    elif args.local_dataset_template:
        with open(args.local_dataset_template, "r") as f:
            rendered_dataset = json.load(f)

    else:
        if mode == "score":
            dataset_template_api = DatasetTemplateApi(
                aoa_client=aoa_client, show_archived=False
            )

            available_templates = list(dataset_template_api)
            template = None
            if not args.local_dataset_template and not args.dataset_template_id:
                template = _select_dataset_template(available_templates)
                if not template:
                    logging.error("No dataset templates found in project")
                    exit(1)

            elif args.dataset_template_id:
                template_exists = next(
                    (
                        True
                        for i in available_templates
                        if i["id"] == args.dataset_template_id
                    ),
                    False,
                )
                if not template_exists:
                    print(
                        "Dataset template not found. Please select one from the list"
                        " below or press Ctrl+C to quit."
                    )
                template = (
                    dataset_template_api.find_by_id(args.dataset_template_id)
                    if template_exists
                    else _select_dataset_template(available_templates)
                )

            dataset_template_id = template["id"]

            rendered_dataset = dataset_template_api.render(id=dataset_template_id)

        else:
            template_api = DatasetTemplateApi(
                aoa_client=aoa_client, show_archived=False
            )
            dataset_api = DatasetApi(aoa_client=aoa_client, show_archived=False)
            dataset = None

            if not args.local_dataset and not args.dataset_id:
                available_templates = list(template_api)
                template = _select_dataset_template(available_templates)
                if not template:
                    logging.error("No dataset template found in project")
                    exit(1)

                from aoa.api.api_iterator import ApiIterator
                from functools import partial

                available_datasets = list(
                    ApiIterator(
                        api_func=partial(
                            dataset_api.find_by_dataset_template_id,
                            template["id"],
                            archived=False,
                        ),
                        entities="datasets",
                    )
                )
                dataset = _select_dataset(available_datasets)
                if not dataset:
                    logging.error("No datasets found in project")
                    exit(1)

            elif args.dataset_id:
                dataset = dataset_api.find_by_id(args.dataset_id)
                if not dataset:
                    print(
                        "Dataset not found. Please select one from the list below or"
                        " press Ctrl+C to quit."
                    )
                    exit(1)

            dataset_id = dataset["id"]

            rendered_dataset = dataset_api.render(dataset_id)

    connection_id = _select_connection(args.connection if args.connection else None)

    print("")
    print("To execute the same command again run:")
    if mode != "score":
        if args.local_dataset:
            print(
                f"tmo run -m {mode} -id {model_id} -c {connection_id} -ld"
                f" {args.local_dataset}"
            )
        else:
            print(
                f"tmo run -m {mode} -id {model_id} -d {dataset_id} -c {connection_id}"
            )
    else:
        if args.local_dataset_template:
            print(
                f"tmo run -m {mode} -id {model_id} -c {connection_id} -lt"
                f" {args.local_dataset_template}"
            )
        else:
            print(
                f"tmo run -m {mode} -id {model_id} -t {dataset_template_id} -c"
                f" {connection_id}"
            )
    print("")
    print("")

    if mode == "train":
        trainer = TrainModel(repo_manager)
        trainer.train_model_local(
            model_id, rendered_dataset=rendered_dataset, base_path=base_path
        )
    elif mode == "evaluate":
        evaluator = EvaluateModel(repo_manager)
        evaluator.evaluate_model_local(
            model_id, rendered_dataset=rendered_dataset, base_path=base_path
        )
    elif mode == "score":
        scorer = ScoreModel(repo_manager)
        scorer.batch_score_model_local(
            model_id, rendered_dataset=rendered_dataset, base_path=base_path
        )
    else:
        logging.error("Unsupported mode used: " + mode)
        exit(1)


def add_task(args, repo_manager, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    if not validate_cwd_valid():
        logging.error(
            "Current working directory must the root of a model repository to execute"
            " this command"
        )
        exit(1)

    import tempfile

    temp_path = tempfile.TemporaryDirectory()
    logging.info("Using this catalog from Git repo {}".format(args.template_url))
    repo_manager.clone_repository(args.template_url, temp_path.name, args.branch)

    task = repo_manager.get_task(source_path=temp_path.name)
    if len(task) == 0:
        logging.error(
            "Repo doesn't contain any valid tasks, use different repo as a model"
            " catalog"
        )
        exit(1)

    task_dir = task

    repo_manager.add_task(task=task_dir, base_path=base_path)
    print("Creating task structure")


def run_task(args, repo_manager, aoa_client, **kwargs):
    from aoa import RunTask

    if args.cwd:
        set_cwd(args.cwd)

    current_project = get_current_project(repo_manager, aoa_client, True)
    if not current_project:
        logging.error(
            "Current working directory must the root of a model repository to execute"
            " this command"
        )
        exit(1)

    aoa_client.set_project_id(current_project["id"])

    def _select_connection(connection=None):
        from argparse import Namespace

        return activate_connection(Namespace(cwd=None, connection=connection))

    connection_id = _select_connection(args.connection if args.connection else None)

    runner = RunTask(repo_manager)

    if not args.function_name:
        func_name = runner.run_task_local(base_path=base_path)
    else:
        func_name = runner.run_task_local(base_path=base_path, func=args.function_name)

    print("")
    print("To execute the same command again run:")
    print(f"tmo task run -f {func_name} -c {connection_id}")
    print("")
    print("")


def list_resources(args, repo_manager, aoa_client, **kwargs):
    from aoa import TrainModel
    from aoa.api.model_api import ModelApi
    from aoa import DatasetApi
    from aoa import DatasetTemplateApi

    if args.cwd:
        set_cwd(args.cwd)

    if args.projects:
        list_and_select_projects(repo_manager, aoa_client, True, False)
        exit(0)

    if args.models:
        project = get_current_project(repo_manager, aoa_client)
        if project:
            aoa_client.set_project_id(project["id"])
        else:
            project = list_and_select_projects(repo_manager, aoa_client, False, True)

        model_api = ModelApi(aoa_client, show_archived=False)
        print_underscored("List of models for project {}:".format(project["name"]))
        if not len(model_api) > 0:
            print("No models were found")
        for i, model in enumerate(model_api):
            print(
                "[{}] Id: ({}) Type: ({}) {}".format(
                    i, model["id"], model["source"], model["name"]
                )
            )

    if args.local_models:
        local_models = TrainModel.get_model_folders(model_catalog, True)
        print_underscored("List of local models:")
        if not len(local_models) > 0:
            print("No local models were found")
        for local_model in local_models:
            print(
                "[{}] {} (Git: {})".format(
                    local_model,
                    local_models[local_model]["name"],
                    local_models[local_model]["id"],
                )
            )

    if args.templates:
        project = get_current_project(repo_manager, aoa_client)
        if project:
            aoa_client.set_project_id(project["id"])
        else:
            project = list_and_select_projects(repo_manager, aoa_client, False, True)

        template_api = DatasetTemplateApi(aoa_client=aoa_client, show_archived=False)
        print_underscored(
            "List of dataset templates for project {}:".format(project["name"])
        )
        if not len(template_api) > 0:
            print("No dataset templates were found")
        for i, template in enumerate(template_api):
            print("[{}] ({}) {}".format(i, template["id"], template["name"]))

    if args.datasets:
        project = get_current_project(repo_manager, aoa_client)
        if project:
            aoa_client.set_project_id(project["id"])
        else:
            project = list_and_select_projects(repo_manager, aoa_client, False, True)

        dataset_api = DatasetApi(aoa_client=aoa_client, show_archived=False)
        print_underscored("List of datasets for project {}:".format(project["name"]))
        if not len(dataset_api) > 0:
            print("No datasets were found")
        for i, dataset in enumerate(dataset_api):
            print("[{}] ({}) {}".format(i, dataset["id"], dataset["name"]))

    if args.connections:
        from argparse import Namespace

        list_connections(Namespace(cwd=None, connections=None))

    elif (
        not args.projects
        and not args.models
        and not args.local_models
        and not args.templates
        and not args.datasets
        and not args.connections
    ):
        logging.error("Invalid object selection...")
        kwargs.get("parent_parser").print_help()
        exit(1)


def clone(args, repo_manager, aoa_client, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    if not args.project_id:
        project = list_and_select_projects(repo_manager, aoa_client, False, True)
    else:
        from aoa import ProjectApi

        project_api = ProjectApi(aoa_client=aoa_client, show_archived=False)
        try:
            project = project_api.find_by_id(args.project_id)
        except Exception:
            project = None
        if not project:
            print(
                "Project not found. Please select one from the list below or press"
                " Ctrl+C to quit."
            )
            project = list_and_select_projects(repo_manager, aoa_client, False, True)

    if args.path:
        path = args.path
    else:
        path = os.path.join(base_path, project["name"])

    project_git = project["gitRepositoryUrl"]

    repo_manager.clone_repository(project_git, path, project.get("branch", "master"))

    config = {"project_id": project["id"]}
    repo_manager.write_repo_config(config, path)
    print("Project cloned at {}".format(path))


def list_connections(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    try:
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        connections_file = "{}/connections.yaml".format(config_dir)
        with open(connections_file, "r") as f:
            connections = yaml.safe_load(f)
    except:
        logging.error("No connections file found. Please add them first.")
        exit(1)

    if "connections" in connections and len(connections["connections"]) > 0:
        print_underscored("List of local connections:")
        for i, connection in enumerate(connections["connections"]):
            print(
                "[{}] ({}) Name: {} Username: {} Host: {} Database: {}".format(
                    i,
                    connection["id"],
                    connection["name"],
                    connection["username"],
                    connection["host"],
                    connection.get("database", ""),
                )
            )
    else:
        logging.error("No connections found in file. Please add them first.")


def add_connections(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    try:
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        connections_file = "{}/connections.yaml".format(config_dir)
        with open(connections_file, "r") as f:
            connections_dict = yaml.safe_load(f)
            if connections_dict is None:
                connections_dict = {}
    except:
        logging.info("No connections file found, creating new one...")
        connections_dict = {}

    connections = (
        connections_dict["connections"]
        if connections_dict
        and "connections" in connections_dict
        and len(connections_dict["connections"]) > 0
        else []
    )

    if args.name and args.username and args.password and args.host:
        name = args.name
        username = args.username
        password = args.password
        host = args.host
        database = args.database
        val_db = args.val_db
        byom_db = args.byom_db
        logmech = args.logmech

    elif not (args.username or args.password or args.name or args.host):
        name = input_string(name="name", tooltip="Type the connection name")
        username = input_string(name="username", tooltip="Type the connection username")
        password = input_string(
            name="password",
            tooltip="Type the connection password (will not show)",
            password=True,
        )
        host = input_string(name="host", tooltip="Type the connection host")
        byom_db = input_string(
            name="BYOM database", required=True, tooltip="The BYOM installation to use"
        )
        val_db = input_string(
            name="VAL database ", required=True, tooltip="The VAL installation to use"
        )
        database = input_string(
            name="database",
            required=False,
            tooltip="Type the default database (optional)",
        )
        logmech = input_string(
            name="logmech", tooltip="The Logmech to use (TD2, LDAP or TDNEGO)"
        )

    else:
        logging.error("Invalid arguments...")
        args.parent_parser.print_help()
        exit(1)

    import uuid

    connection_id = str(uuid.uuid4())
    encrypted_password = crypto.td_encrypt_password(
        password=password,
        pass_key_filename="{}/{}.key".format(config_dir, connection_id),
        enc_pass_filename="{}/{}.pass".format(config_dir, connection_id),
    )

    connections.append({
        "id": connection_id,
        "name": name,
        "username": username,
        "password": encrypted_password,
        "host": host,
        "logmech": logmech,
        "database": database,
        "val_db": val_db,
        "byom_db": byom_db,
    })
    connections_dict["connections"] = connections

    try:
        connections_file = "{}/connections.yaml".format(config_dir)
        with open(connections_file, "w+") as f:
            yaml.safe_dump(connections_dict, f, default_flow_style=False)
    except Exception as ex:
        logging.error("Can't save connections file: {}".format(ex))
        exit(1)


def remove_connections(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    def _check_connection_exists(id, obj):
        for connection in obj:
            if "id" in connection and connection["id"] == id:
                return True
        return False

    def _remove_connection(id, obj):
        if os.path.exists("{}/{}.key".format(config_dir, id)):
            os.remove("{}/{}.key".format(config_dir, id))
        if os.path.exists("{}/{}.pass".format(config_dir, id)):
            os.remove("{}/{}.pass".format(config_dir, id))
        result = []
        for item in obj:
            if item["id"] != id:
                result.append(item)
        return result

    try:
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        connections_file = "{}/connections.yaml".format(config_dir)
        with open(connections_file, "r") as f:
            connections_dict = yaml.safe_load(f)
            if connections_dict is None or not (
                "connections" in connections_dict
                and len(connections_dict["connections"]) > 0
            ):
                logging.info("No connections defined, nothing to remove.")
    except:
        logging.info("No connections file found, nothing to remove.")
        exit(0)

    connections = (
        connections_dict["connections"]
        if "connections" in connections_dict
        and len(connections_dict["connections"]) > 0
        else []
    )
    connection = None
    if args.connection:
        id = args.connection
        if not _check_connection_exists(id, connections):
            logging.info("Connection does not exists, exiting.")
            exit(1)
    else:
        name = input_select(
            "connection",
            [item["name"] for item in connections],
            "List of local connections",
        )
        connection = next(
            (
                connections[i]["id"]
                for i, item in enumerate(connections)
                if connections[i]["name"] == name
            ),
            None,
        )

    connections = _remove_connection(connection, connections)
    connections_dict["connections"] = connections

    try:
        connections_file = "{}/connections.yaml".format(config_dir)
        with open(connections_file, "w+") as f:
            yaml.safe_dump(connections_dict, f, default_flow_style=False)
    except Exception as ex:
        logging.error("Can't save connections file: {}".format(ex))
        exit(1)


def export_connection(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    def _check_connection_exists(id, obj):
        for connection in obj:
            if "id" in connection and connection["id"] == id:
                return True
        return False

    try:
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        connections_file = "{}/connections.yaml".format(config_dir)
        with open(connections_file, "r") as f:
            connections_dict = yaml.safe_load(f)
            if connections_dict is None or not (
                "connections" in connections_dict
                and len(connections_dict["connections"]) > 0
            ):
                logging.info("No connections defined, Please add them first.")
    except:
        logging.info("No connections file found, Please add them first.")
        exit(0)

    connection = None
    connections = (
        connections_dict["connections"]
        if "connections" in connections_dict
        and len(connections_dict["connections"]) > 0
        else []
    )
    if args.connection:
        id = args.connection
        if not _check_connection_exists(id, connections):
            logging.info("Connection does not exists, exiting.")
            exit(1)
    else:
        name = input_select(
            "connection",
            [item["name"] for item in connections],
            "List of local connections",
        )
        connection = next(
            (
                connections[i]
                for i, item in enumerate(connections)
                if connections[i]["name"] == name
            ),
            None,
        )

    print("\nCopy the below command and execute in your terminal: \n")
    print(
        'export AOA_CONN_USERNAME="{}" && \\\n'.format(
            bash_escape(connection["username"])
        )
        + 'export AOA_CONN_PASSWORD="{}" && \\\n'.format(
            bash_escape(connection["password"])
        )
        + 'export AOA_CONN_HOST="{}" && \\\n'.format(bash_escape(connection["host"]))
        + 'export AOA_CONN_LOG_MECH="{}" && \\\n'.format(
            bash_escape(connection["logmech"])
        )
        + 'export AOA_CONN_DATABASE="{}" && \\\n'.format(
            bash_escape(connection.get("database", ""))
        )
        + 'export AOA_VAL_INSTALL_DB="{}" && \\\n'.format(
            bash_escape(connection.get("val_db", "VAL"))
        )
        + 'export AOA_BYOM_INSTALL_DB="{}" && \\\n'.format(
            bash_escape(connection.get("byom_db", "BYOM"))
        )
    )


def activate_connection(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    def _check_connection_exists(id, obj):
        for connection in obj:
            if "id" in connection and connection["id"] == id:
                return True
        return False

    def _activate_connection(id, obj):
        for connx in obj:
            if "id" in connx and connx["id"] == id:
                os.environ["AOA_CONN_USERNAME"] = connx["username"]
                os.environ["AOA_CONN_PASSWORD"] = connx["password"]
                os.environ["AOA_CONN_HOST"] = connx["host"]
                os.environ["AOA_CONN_LOG_MECH"] = connx["logmech"]
                os.environ["AOA_CONN_DATABASE"] = connx.get("database", "")
                os.environ["AOA_VAL_INSTALL_DB"] = connx.get("val_db", "VAL")
                os.environ["AOA_BYOM_INSTALL_DB"] = connx.get("ml_db", "MLDB")

                return True
        return False

    try:
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        connections_file = "{}/connections.yaml".format(config_dir)
        with open(connections_file, "r") as f:
            connections_dict = yaml.safe_load(f)
            if connections_dict is None:
                connections_dict = {}
    except:
        logging.info("No connections file found, you must create a connection first")
        exit(1)

    connections = (
        connections_dict["connections"]
        if "connections" in connections_dict
        and len(connections_dict["connections"]) > 0
        else []
    )
    if args.connection:
        connection = args.connection
    elif kwargs.get("connection"):
        connection = kwargs.get("connection")
    else:
        if len(connections) == 1:
            print(
                "Automatic connection selection as only one available: {}".format(
                    connections[0]["name"]
                )
            )
            connection = connections[0]["id"]
        else:
            selected_connection_value = input_select(
                "connection",
                [item["name"] for item in connections],
                "Available connections:",
            )
            connection = next(
                (
                    connections[i]["id"]
                    for i, item in enumerate(connections)
                    if connections[i]["name"] == selected_connection_value
                ),
                None,
            )

    if _check_connection_exists(connection, connections):
        _activate_connection(connection, connections)
    else:
        logging.error("The specified connection was not found.")
        exit(1)

    return connection


def test_connection(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    def _test_connection():
        from aoa.util import aoa_create_context
        from aoa.util.connections import execute_sql

        aoa_create_context()
        test_query = "SEL infodata FROM dbc.dbcinfo WHERE infokey = 'VERSION'"
        logging.debug(test_query)
        result = execute_sql(test_query).fetchall()
        return result[0][0]

    from argparse import Namespace

    activate_connection(
        Namespace(cwd=None, connection=args.connection if args.connection else None)
    )

    try:
        ver = _test_connection()
        print("Connection successful, Vantage version: {}".format(ver))
    except Exception as ex:
        logging.error("Failed to test connecton: {}".format(ex))
        logging.error("Please use 'tmo connection add' to add working connection")
        exit(1)


def create_byom_table(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    from aoa.stats import store
    from aoa.util.connections import execute_sql

    ct_query = store.byom_query.format(args.name)
    print(ct_query)

    if args.execute_ddl:
        from argparse import Namespace

        activate_connection(Namespace(cwd=None, connection=None))
        from aoa.util import aoa_create_context

        try:
            aoa_create_context()
            logging.debug(ct_query)
            execute_sql(ct_query)
        except Exception as ex:
            raise Exception("Could not create BYOM table: {}".format(ex)) from ex

        print("BYOM table {} created successfully".format(args.name))
    else:
        print(
            "Execution not requested, just showing DDL. To execute this DDL, add -e to"
            " the command."
        )


def compute_stats(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    from argparse import Namespace

    activate_connection(Namespace(cwd=None, connection=None))

    from aoa.util import aoa_create_context
    from aoa.stats import stats, store
    from teradataml import DataFrame

    try:
        aoa_create_context()

        df = DataFrame.from_query(f"SEL * FROM {args.source_table}")

        columns = list(map(str.strip, map(str.lower, args.columns.split(","))))

        if args.feature_type == "categorical":
            fs = stats.compute_categorical_stats(df, columns)
            store.save_feature_stats(
                features_table=args.metadata_table, feature_type="categorical", stats=fs
            )

        elif args.feature_type == "continuous":
            fs = stats.compute_continuous_stats(df, columns)
            store.save_feature_stats(
                features_table=args.metadata_table, feature_type="continuous", stats=fs
            )
    except Exception as ex:
        raise Exception("Could not compute feature stats: {}".format(ex)) from ex

    logging.info("Statistics computed successfully")


def list_stats(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    from argparse import Namespace

    activate_connection(Namespace(cwd=None, connection=None))
    from aoa.util import aoa_create_context
    from aoa.stats import store

    try:
        aoa_create_context()
        stats_summary = store.get_feature_stats_summary(args.metadata_table)
        logging.debug(stats_summary)

        if len(stats_summary) < 0:
            raise Exception(
                "No statistics stored in the table {}".format(args.metadata_table)
            )

        width = len(max(list(stats_summary.keys()), key=len))
        print_underscored("List of stored feature stats:")
        for v in stats_summary:
            print(
                "{feature: <{width}} | {type}".format(
                    feature=v, width=width, type=stats_summary[v]
                )
            )
    except Exception as ex:
        raise Exception("Could not list feature stats: {}".format(ex)) from ex


def create_stats_table(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)

    from aoa.stats.store import get_features_stats_metadata_ct_query
    from aoa.util.connections import execute_sql

    ct_query = get_features_stats_metadata_ct_query(args.metadata_table)
    print(ct_query)

    if args.execute_ddl:
        from argparse import Namespace

        activate_connection(Namespace(cwd=None, connection=None))
        from aoa.util import aoa_create_context

        try:
            aoa_create_context()
            logging.debug(ct_query)
            execute_sql(ct_query)
        except Exception as ex:
            raise Exception("Could not create statistics: {}".format(ex)) from ex

        print("Statistics table {} created successfully".format(args.metadata_table))
    else:
        print(
            "Execution not requested, just showing DDL. To execute this DDL, add -e to"
            " the command."
        )


def import_stats(args, **kwargs):
    if args.cwd:
        set_cwd(args.cwd)
    if args.show_example:
        print("""
            {
                "age": {
                    "type": "continuous",
                    "edges": [21.0, 27.0, 33.0, 39.0, 45.0, 51.0, 57.0, 63.0, 69.0, 75.0, 81.0]
                }, 
                "hasdiabetes": {
                    "type": "categorical",
                    "frequencies": ["0", "1"]
                }
            }
        """)
        exit(0)

    from argparse import Namespace

    activate_connection(Namespace(cwd=None, connection=None))
    from aoa.util import aoa_create_context
    from aoa.stats import store
    import json

    try:
        aoa_create_context()
        json_file = open(args.statistics_file)
        stats = json.load(json_file)

        categorical_features = {
            x: stats[x] for x in stats if stats[x]["type"] == "categorical"
        }
        store.save_feature_stats(
            args.metadata_table, "categorical", categorical_features
        )
        continuous_features = {
            x: stats[x] for x in stats if stats[x]["type"] == "continuous"
        }
        store.save_feature_stats(args.metadata_table, "continuous", continuous_features)

    except Exception as ex:
        raise Exception("Could not list feature stats: {}".format(ex)) from ex

    print(
        "Successfully saved statistics in {} features table".format(args.metadata_table)
    )


def doctor(args, repo_manager, aoa_client, **kwargs):
    print("Testing ModelOps service configuration")

    from aoa import ProjectApi

    try:
        project_api = ProjectApi(aoa_client=aoa_client, show_archived=False)
        projects = list(project_api)
        if len(projects) > 0:
            print(f"ModelOps service configured, number of projects: {len(projects)}")
        else:
            raise
    except Exception as ex:
        logging.error(
            "Failed to connect to ModelOps or find any project. \n"
            + "Please (re)copy the CLI configuration from ModelOps UI -> Session"
            " Details -> CLI Config"
        )

    print("")

    print("Testing Vantage connections")
    args.connection = None
    test_connection(args)


def set_cwd(path):
    import os

    if not os.path.exists(path):
        logging.error("The specified path does not exist... exiting")
        exit(1)
    os.chdir(path)
    global base_path, model_catalog
    base_path = os.path.abspath(os.getcwd())
    model_catalog = os.path.join(base_path, "model_definitions/")


def print_help(args, **kwargs):
    from aoa import __version__

    if args.version:
        print("{}".format(__version__))
    else:
        kwargs.get("parent_parser").print_help()


def print_underscored(message):
    print(message)
    print("-" * len(message))


def list_and_select_projects(
    repo_manager, aoa_client, as_list=False, check_config=False
):
    from aoa import ProjectApi

    project_api = ProjectApi(aoa_client=aoa_client, show_archived=False)
    projects = list(project_api)

    print_underscored("List of projects:" if as_list else "Available projects:")
    if not len(projects) > 0:
        print("No projects were found")
    for ix, item in enumerate(projects):
        print("[{}] ({}) {}".format(ix, item["id"], item["name"]))
    if as_list:
        return

    current_project = get_current_project(repo_manager, aoa_client, check_config)
    current_project = current_project if current_project else ""
    current_index = "none"
    for ix, item in enumerate(projects):
        if "id" in current_project and current_project["id"] == item["id"]:
            current_index = ix
    tmp_index = input(
        "Select project by index (current selection: {}): ".format(current_index)
    )
    print("")

    if (
        (not tmp_index.isnumeric() or int(tmp_index) >= len(projects))
        and not tmp_index == ""
    ) or (tmp_index == "" and current_index == "none"):
        print(
            "Wrong selection, please try again by selecting the index number on the"
            " first column."
        )
        print("")
        return list_and_select_projects(repo_manager, aoa_client, as_list, check_config)

    if tmp_index == "":
        tmp_index = current_index

    selected_project = projects[int(tmp_index)]

    if aoa_client:
        aoa_client.set_project_id(selected_project["id"])

    return selected_project


def get_current_project(repo_manager, aoa_client, check_repo_conf=False):
    from aoa import ProjectApi

    if not validate_cwd_valid():
        return

    repo_conf = repo_manager.read_repo_config()
    current_project_id = (
        repo_conf["project_id"] if repo_conf and "project_id" in repo_conf else ""
    )
    project_api = ProjectApi(aoa_client=aoa_client, show_archived=False)
    current_project = project_api.find_by_id(current_project_id)

    if current_project:
        return current_project
    elif not current_project and check_repo_conf:
        print("The repository is not linked to a ModelOps project.")
        print("Please execute 'tmo link'")
        print("")

    return


def link_repo(repo_manager, aoa_client, **kwargs):
    current_project = list_and_select_projects(repo_manager, aoa_client, False, False)
    config = {"project_id": current_project["id"]}
    repo_manager.write_repo_config(config)
    print("Repo linked to Project.")


def handle_generic_error(ex: Exception, debug: bool):
    if debug:
        logging.exception("An error occurred, printing stack trace output.")
    else:
        logging.error("An error occurred: {}".format(ex))
        exit(1)


def main():
    def __remove_cached_token():
        if os.path.exists(AoaClient.DEFAULT_TOKEN_CACHE_FILE_PATH):
            os.remove(AoaClient.DEFAULT_TOKEN_CACHE_FILE_PATH)

    args = None
    try:
        parent_parser = argparse.ArgumentParser(description="TMO CLI")
        parent_parser.add_argument(
            "--debug", action="store_true", help="Enable debug logging"
        )
        parent_parser.add_argument(
            "--version", action="store_true", help="Display the version of this tool"
        )
        parent_parser.set_defaults(func=print_help)

        subparsers = parent_parser.add_subparsers(
            title="actions", description="valid actions", dest="command"
        )

        common_parser = argparse.ArgumentParser(add_help=False)
        common_parser.add_argument(
            "--debug", action="store_true", help="Enable debug logging"
        )
        common_parser.add_argument("-cwd", "--cwd", type=str, help=argparse.SUPPRESS)
        common_parser.add_argument(
            "--test", action="store_true", help=argparse.SUPPRESS
        )

        parser_list = subparsers.add_parser(
            "list",
            help="List projects, models, local models or datasets",
            parents=[common_parser],
        )
        parser_list.add_argument(
            "-p", "--projects", action="store_true", help="List projects"
        )
        parser_list.add_argument(
            "-m",
            "--models",
            action="store_true",
            help="List registered models (committed / pushed)",
        )
        parser_list.add_argument(
            "-lm",
            "--local-models",
            action="store_true",
            help=(
                "List local models. Includes registered and non-registered"
                " (non-committed / non-pushed)"
            ),
        )
        parser_list.add_argument(
            "-t", "--templates", action="store_true", help="List dataset templates"
        )
        parser_list.add_argument(
            "-d", "--datasets", action="store_true", help="List datasets"
        )
        parser_list.add_argument(
            "-c", "--connections", action="store_true", help="List local connections"
        )
        parser_list.set_defaults(func=list_resources)

        parser_add = subparsers.add_parser(
            "add", help="Add model to working dir", parents=[common_parser]
        )
        parser_add.add_argument(
            "-t",
            "--template-url",
            type=str,
            help="Git URL for template repository",
            required=True,
        )
        parser_add.add_argument(
            "-b",
            "--branch",
            type=str,
            default="main",
            help="Git branch to pull templates (default is main)",
            required=True,
        )
        parser_add.set_defaults(func=add_model)

        parser_run = subparsers.add_parser(
            "run", help="Train and Evaluate model locally", parents=[common_parser]
        )
        parser_run.add_argument("-id", "--model-id", type=str, help="Id of model")
        parser_run.add_argument(
            "-m", "--mode", type=str, help="Mode (train or evaluate)"
        )
        parser_run.add_argument("-d", "--dataset-id", type=str, help="Remote datasetId")
        parser_run.add_argument(
            "-t", "--dataset-template-id", type=str, help="Remote datasetTemplateId"
        )
        parser_run.add_argument(
            "-ld",
            "--local-dataset",
            type=str,
            help="Path to local dataset metadata file",
        )
        parser_run.add_argument(
            "-lt",
            "--local-dataset-template",
            type=str,
            help="Path to local dataset template metadata file",
        )
        parser_run.add_argument(
            "-c", "--connection", type=str, help="Local connection id"
        )
        parser_run.set_defaults(func=run_model)

        parser_init = subparsers.add_parser(
            "init",
            help="Initialize model directory with basic structure",
            parents=[common_parser],
        )
        parser_init.set_defaults(func=init_model_directory)

        parser_clone = subparsers.add_parser(
            "clone", help="Clone Project Repository", parents=[common_parser]
        )
        parser_clone.add_argument(
            "-id", "--project-id", type=str, help="Id of Project to clone"
        )
        parser_clone.add_argument(
            "-p", "--path", type=str, help="Path to clone repository to"
        )
        parser_clone.set_defaults(func=clone)

        parser_configure = subparsers.add_parser(
            "link", help="Link local repo to project", parents=[common_parser]
        )
        parser_configure.set_defaults(func=link_repo)

        parser_tasks = subparsers.add_parser(
            "task", help="Manage feature engineering tasks"
        )
        subparser_tasks = parser_tasks.add_subparsers(
            title="actions", description="valid actions", dest="command"
        )

        subparser_add_task = subparser_tasks.add_parser(
            "add",
            help="Add feature engineering task to working dir",
            parents=[common_parser],
        )
        subparser_add_task.add_argument(
            "-t",
            "--template-url",
            type=str,
            help="Git URL for template repository",
            required=True,
        )
        subparser_add_task.add_argument(
            "-b",
            "--branch",
            type=str,
            default="main",
            help="Git branch to pull task (default is main)",
            required=True,
        )
        subparser_add_task.set_defaults(func=add_task)

        subparser_run_task = subparser_tasks.add_parser(
            "run", help="Run feature engineering tasks locally", parents=[common_parser]
        )

        subparser_run_task.add_argument(
            "-c", "--connection", type=str, help="Local connection id"
        )
        subparser_run_task.add_argument(
            "-f", "--function-name", type=str, help="Task function name"
        )
        subparser_run_task.set_defaults(func=run_task)
        parser_connections = subparsers.add_parser(
            "connection", help="Manage local connections"
        )
        subparser_connections = parser_connections.add_subparsers(
            title="actions", description="valid actions", dest="command"
        )
        subparser_list_connections = subparser_connections.add_parser(
            "list", help="List all local connections", parents=[common_parser]
        )
        subparser_list_connections.set_defaults(func=list_connections)
        subparser_add_connections = subparser_connections.add_parser(
            "add", help="Add a local connection", parents=[common_parser]
        )
        subparser_add_connections.add_argument(
            "-n", "--name", type=str, help="Connection name"
        )
        subparser_add_connections.add_argument(
            "-H", "--host", type=str, help="Connection host"
        )
        subparser_add_connections.add_argument(
            "-u", "--username", type=str, help="Connection username"
        )
        subparser_add_connections.add_argument(
            "-p", "--password", type=str, help="Connection password"
        )
        subparser_add_connections.add_argument(
            "-d", "--database", type=str, help="Connection default database"
        )
        subparser_add_connections.add_argument(
            "--val-db",
            type=str,
            default="val",
            help="Configure VAL installation db (default is val)",
        )
        subparser_add_connections.add_argument(
            "--byom-db",
            type=str,
            default="mldb",
            help="Configure BYOM installation db (default is mldb)",
        )
        subparser_add_connections.add_argument(
            "-l", "--logmech", type=str, help="Logmech to use"
        )
        subparser_add_connections.set_defaults(
            func=add_connections, parent_parser=subparser_add_connections
        )
        subparser_remove_connections = subparser_connections.add_parser(
            "remove", help="Remove a local connection", parents=[common_parser]
        )
        subparser_remove_connections.add_argument(
            "-c", "--connection", type=str, help="Local connection id"
        )
        subparser_remove_connections.set_defaults(func=remove_connections)
        subparser_export_connections = subparser_connections.add_parser(
            "export",
            help="Export a local connection to be used as a shell script",
            parents=[common_parser],
        )
        subparser_export_connections.add_argument(
            "-c", "--connection", type=str, help="Local connection id"
        )
        subparser_export_connections.set_defaults(func=export_connection)
        subparser_test_connections = subparser_connections.add_parser(
            "test", help="Test a local connection", parents=[common_parser]
        )
        subparser_test_connections.add_argument(
            "-c", "--connection", type=str, help="Local connection id"
        )
        subparser_test_connections.set_defaults(func=test_connection)

        subparser_create_byom_table = subparser_connections.add_parser(
            "create-byom-table",
            help="Create a table to store BYOM models",
            parents=[common_parser],
        )
        subparser_create_byom_table.add_argument(
            "-n",
            "--name",
            type=str,
            default="tmo_byom_models",
            help="Name of BYOM table (default is tmo_byom_models)",
        )
        subparser_create_byom_table.add_argument(
            "-e",
            "--execute-ddl",
            action="store_true",
            help="Execute CREATE TABLE DDL, not just generate it",
        )
        subparser_create_byom_table.set_defaults(func=create_byom_table)

        parser_features = subparsers.add_parser(
            "feature", help="Manage feature statistics"
        )
        subparser_feature = parser_features.add_subparsers(
            title="action", description="valid actions", dest="command"
        )
        subparser_compute_stats = subparser_feature.add_parser(
            "compute-stats", help="Compute feature statistics", parents=[common_parser]
        )
        subparser_compute_stats.add_argument(
            "-s",
            "--source-table",
            type=str,
            help="Feature source table/view",
            required=True,
        )
        subparser_compute_stats.add_argument(
            "-m",
            "--metadata-table",
            type=str,
            help="Metadata table for feature stats, including database name",
            required=True,
        )
        subparser_compute_stats.add_argument(
            "-t",
            "--feature-type",
            choices=["continuous", "categorical"],
            default="continuous",
            help="Feature type: continuous or categorical (default is continuous)",
        )
        subparser_compute_stats.add_argument(
            "-c", "--columns", type=str, help="List of feature columns", required=True
        )
        subparser_compute_stats.set_defaults(
            func=compute_stats, parent_parser=subparser_compute_stats
        )
        subparser_list_stats = subparser_feature.add_parser(
            "list-stats", help="List available statistics", parents=[common_parser]
        )
        subparser_list_stats.add_argument(
            "-m",
            "--metadata-table",
            type=str,
            help="Metadata table for feature stats, including database name",
            required=True,
        )
        subparser_list_stats.set_defaults(
            func=list_stats, parent_parser=subparser_list_stats
        )
        subparser_create_stats_table = subparser_feature.add_parser(
            "create-stats-table",
            help="Create statistics table",
            parents=[common_parser],
        )
        subparser_create_stats_table.add_argument(
            "-m",
            "--metadata-table",
            type=str,
            help="Metadata table for feature stats, including database name",
            required=True,
        )
        subparser_create_stats_table.add_argument(
            "-e",
            "--execute-ddl",
            action="store_true",
            help="Execute CREATE TABLE DDL, not just generate it",
        )
        subparser_create_stats_table.set_defaults(
            func=create_stats_table, parent_parser=subparser_create_stats_table
        )
        subparser_import_stats = subparser_feature.add_parser(
            "import-stats",
            help="Import column statistics from local JSON file",
            parents=[common_parser],
        )
        subparser_import_stats.add_argument(
            "-m",
            "--metadata-table",
            type=str,
            help="Metadata table for feature stats, including database name",
            required=True,
        )
        subparser_import_stats.add_argument(
            "-f",
            "--statistics-file",
            type=str,
            help="Name of statistics JSON file",
            required=True,
        )
        subparser_import_stats.add_argument(
            "-s",
            "--show-example",
            action="store_true",
            help="Show file structure example and exit",
        )
        subparser_import_stats.set_defaults(
            func=import_stats, parent_parser=subparser_import_stats
        )

        parser_doctor = subparsers.add_parser(
            "doctor", help="Diagnose configuration issues", parents=[common_parser]
        )
        parser_doctor.set_defaults(func=doctor)

        args = parent_parser.parse_args()

        logging_level = logging.DEBUG if args.debug else logging.INFO
        logging.basicConfig(
            format="%(message)s", stream=sys.stdout, level=logging_level
        )
        print("")

        # Certain functions don't need an API connection, can be called without initializing a client and repo manager
        local_functions = [
            print_help,
            list_connections,
            remove_connections,
            add_connections,
            test_connection,
            export_connection,
            create_byom_table,
            create_stats_table,
            import_stats,
            list_stats,
            compute_stats,
        ]
        if args.command and args.func not in local_functions:
            from aoa import RepoManager
            from aoa import AoaClient

            repo_manager = RepoManager(base_path)
            aoa_client = AoaClient()

            args.func(
                repo_manager=repo_manager,
                aoa_client=aoa_client,
                args=args,
                parent_parser=parent_parser,
            )
        else:
            args.func(args=args, parent_parser=parent_parser)

    except requests.exceptions.ConnectionError as ce:
        logging.error(
            COLOR["WARN"]
            + "Could not connect to ModelOps API. "
            + COLOR["END"]
            + "\n"
            + "Please verify you are connected to network and have access to ModelOps"
            " UI.\n"
            + "If you are connected and have access, "
            + "try (re)copy the CLI configuration from ModelOps UI -> Session Details"
            " -> CLI Config.\n\n"
            + "For more details, run with --debug.\n"
        )

        if args.debug:
            logging.exception(
                COLOR["WARN"]
                + "An error occurred, printing stack trace output."
                + COLOR["END"]
            )

        exit(1)

    except oauthlib.oauth2.rfc6749.errors.InvalidGrantError as ge:
        if ge.description in ["Token is not active", "Session not active"]:
            logging.error(
                ge.description
                + "\n"
                + COLOR["WARN"]
                + "Clearing the token cache. Please re-run cmd and login again."
                + COLOR["END"]
                + "\n"
            )

            __remove_cached_token()

            exit(1)
        else:
            handle_generic_error(ge, args.debug)

    except oauthlib.oauth2.rfc6749.errors.InvalidTokenError as ge:
        logging.error(
            ge.description
            + "\n"
            + COLOR["WARN"]
            + "Clearing the token cache. Please re-run cmd and login again."
            + COLOR["END"]
            + "\n"
        )

        __remove_cached_token()

        exit(1)

    except KeyboardInterrupt:
        logging.info("")
        logging.info(
            f"{COLOR['WARN']}Keyboard interrupt... exiting{COLOR['END']}\33[?25h"
        )
        exit(1)

    except Exception as ex:
        handle_generic_error(ex, args.debug)


filename = re.findall('File\s".*\/(.*)"', traceback.format_stack()[0])[0]

if filename in ["aoa"]:
    print(
        COLOR["WARN"],
        "This is an old version of Teradata ModelOps utility, still working, but"
        " consider using tmo command instead of aoa.",
        COLOR["END"],
    )

if __name__ in ["__main__", "aoa", "tmo"]:
    main()



from __future__ import absolute_import
from typing import Dict, List

from aoa.api.iterator_base_api import IteratorBaseApi


class FeatureEngineeringApi(IteratorBaseApi):

    path = "/api/featureEngineeringTasks"
    type = "FEATURE_ENGINEERING"

    def _get_header_params(self):
        # The header for project id is required for the archive/unarchive method from base_api
        header_vars = [
            "AOA-Project-ID",
            "VMO-Project-ID",
            "Content-Type",
            "Accept",
        ]  # AOA-Project-ID kept for backwards compatibility
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.project_id,
            self.json_type,
            self.json_type,
        ]

        return self.generate_params(header_vars, header_vals)

    def import_task(self, import_request: Dict[str, str]):
        """
        import a feature engineering task

        Parameters:
            import_request (dict): request to import model version

        Returns:
            (dict): job
        """
        self.required_params(
            ["artefactImportId", "name", "description", "language", "functionName"],
            import_request,
        )

        return self.aoa_client.post_request(
            path=f"{self.path}/import",
            header_params=self._get_header_params(),
            query_params={},
            body=import_request,
        )

    def run(self, task_id: str, run_request: Dict[str, str]):
        """
        run feature engineering task

        Parameters:
            task_id (str): model id(uuid)
            run_request (dict): request to import model version

        Returns:
            (dict): job
        """
        self.required_params(["automation", "datasetConnectionId"], run_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{task_id}/run",
            header_params=self._get_header_params(),
            query_params={},
            body=run_request,
        )

    def approve(self, task_id: str, comments: str):
        """
        approve a feature engineering task
        :param task_id:  task id
        :param comments: approval comments
        :return:
        """

        approve_request = dict()
        approve_request["comments"] = comments

        return self.aoa_client.post_request(
            path=f"{self.path}/{task_id}/approve",
            header_params=self._get_header_params(),
            query_params={},
            body=approve_request,
        )

    def reject(self, task_id: str, comments: str):
        """
        reject a feature engineering task
        :param task_id:  task id
        :param comments: approval comments
        :return:
        """

        reject_request = dict()
        reject_request["comments"] = comments

        return self.aoa_client.post_request(
            path=f"{self.path}/{task_id}/approve",
            header_params=self._get_header_params(),
            query_params={},
            body=reject_request,
        )

    def deploy(self, task_id: str, deploy_request: dict):
        """
        deploy a trained model
        :param task_id:  fe task id
        :param deploy_request: deployment request
        :return:
        """
        self.required_params(["engineType"], deploy_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{task_id}/deploy",
            header_params=self._get_header_params(),
            query_params={},
            body=deploy_request,
        )

    def retire(self, task_id: str, retire_request: dict):
        """
        retire a trained model
        :param task_id:  fe task id
        :param retire_request: retire request
        :return:
        """
        self.required_params(["deploymentId"], retire_request)

        return self.aoa_client.post_request(
            path=f"{self.path}/{task_id}/retire",
            header_params=self._get_header_params(),
            query_params={},
            body=retire_request,
        )



from teradataml import DataFrame
from .connections import execute_sql_in_context


def store_byom_tmp(
    context, table: str, model_version: str, model_bytes: bytes
) -> DataFrame:
    """
    Store a byom model in a temporary table so we can use it

    :param context: get_context()
    :param table: temp table name to use
    :param model_version: model version
    :param model_bytes: model bytes
    :return: DataFrame of model
    """
    execute_sql_in_context(
        context,
        f"""
        CREATE VOLATILE TABLE {table}(
            model_id VARCHAR(255),
            model BLOB(2097088000)
        ) ON COMMIT PRESERVE ROWS;
        """,
    )

    execute_sql_in_context(
        context,
        f"INSERT INTO {table}(model_id, model) values(?,?)",
        (model_version, model_bytes),
    )

    return DataFrame.from_query(f"""
        SELECT model_id, model FROM {table} 
            WHERE model_id='{model_version}'
        """)


def check_if_table_exists(context, database: str, name: str, table_kind: str) -> bool:
    """
    Check if the table/view/etc exists in the database

    :param context: sqlalchemy context/engine
    :param database: the database
    :param name: the table/view/etc name
    :param table_kind: 'V' for view, 'T' for table etc
    :return: (bool) True if exists, False otherwise
    """
    rs = execute_sql_in_context(
        context,
        f"""
    SELECT * FROM DBC.TABLES WHERE 
        TABLENAME ='{name}' AND 
        tablekind = '{table_kind}' AND 
        databasename='{database}'
    """,
    )

    return rs.rowcount >= 1
